{-|
   A module for converting Pepa models to hydra models
-}

module Language.Pepa.Compile.Hydra
   ( dnamFromPepa
   , addMeasurementSpecs
   , translateTransition
   , ApparentRate
   , DNAMgenParTransitions
   )
where

{- Imported Standard Libraries -}
import Data.List
   ( partition
   , find
   )
{- Imported Local Libraries -}
import Language.Pepa.Syntax
   ( QualifiedName         ( .. )
   , concatQNames
   , prefixQName
   , ShowOrig              ( .. )
   , ParsedComponentId
   , ParsedModel           ( .. )
   , ParsedTrans           ( .. )
   , ParsedAction
   , RateSpec
   , RateIdentifier
   , ParsedRate            ( .. )
   , ParsedRateExp         ( .. )

   , nameOfAction
   ) 
import Language.Hydra.Syntax
   ( DNAMmodelFile         ( .. )
   , DNAMmodel             ( .. )
   , DNAMheader            ( .. )

   , DNAMidentifier

   , DNAMconstantDef       ( .. )
   , DNAMinitialVector
   , DNAMstateVector
   , DNAMstateDesc         ( .. )

   , DNAMmodelTransitions
   , DNAMtransition        ( .. )
   , DNAMcondition         ( .. )
   , cExpOfDNAMcond
   , DNAMspeed             ( .. )
   , DNAMaction


   , DNAMsolutionControl
   , DNAMperformMeasure    ( .. )
   , DNAMpassageMeasure    ( .. )
   , DNAMtransientMeasure  ( .. )
   , DNAMsteadyMeasure     ( .. )
   , DNAMcountMeasure      ( .. )

   , DNAMcassignment       ( .. )
   , DNAMctype             ( .. )
   , DNAMcexp              ( .. )
   , dnamMinimum
   )
import Language.Hydra.Transform
   ( uniquifyTransitions )

import Language.Pepa.Compile.GenericParallel
   ( Successor
   , SuccessorMap
   , GPelement             ( .. )
   , GenericParallel       ( .. )
   , toGenericParallel
   )

import Language.Pepa.Analysis.Analysis
   (  ConcentrationMap
   , getInitialConcentrations
   )
{- End of Imports -}


{-| 
   Sorry for the messy definitions based on this extra constructor
   hopefully it won't be hanging around that long.

   Hmm, well it seems as though it is, essentially then is just to allow
   this particular kind of dnam transitions to be a generic parallel element.
-}
newtype DNAMgenParTransitions = 
   DNAMgenParTransitions (DNAMmodelTransitions ParsedAction ApparentRate)

instance GPelement DNAMgenParTransitions where
   gpIdElement     = fromPepaComponent
   stopElement     = DNAMgenParTransitions []


{-|
   To calculate the apparent rate we require to calculate r_a(P) which is
   essentially the sum of the rates of all ways in which P may perform
   an /a/ action. It cannot mix passive and active so either this value is top
   multiplied by a number or is the sum of a list or reals.
   We must keep the conditions so that we can say whether or not it is really
   possible for that transition to occur.
-}
type RateDivisor = [ ( [ DNAMcondition ], ApparentRate ) ]

{-|
   An apparent rate is marked as either passive or active.
   This allows us to take the minimum of two, where the minimum of
   an active and passive rate is always the active rate.
-}
data RateKind = Active | Passive
                deriving Eq

{-|
    I'd like to move this and 'RateDivisor' 
   (probably parameterised on the type of the condition)
   into 'Language.Pepa.Syntax' since I think it will be useful for translations
   other than into Hydra.
-}
data ApparentRate = 
   SingleRate ParsedRate
 | CompoundRate RateKind ApparentRate ApparentRate RateDivisor RateDivisor

rateKind :: ApparentRate -> RateKind
rateKind (SingleRate RateTop)      = Passive
rateKind (SingleRate _)            = Active
rateKind (CompoundRate rk _ _ _ _) = rk

{-
instance DNAMspeedTarget ApparentRate where
   toDnamSpeed (SingleRate r) = toDnamSpeed r
   toDnamSpeed r              = DNAMrate $ cExprOfApparentRate r
-}

{-|
   Translation from transitions containing apparent rates to
   transitions containing dnam speeds.
-}  
translateTransition :: Bool -- ^ Whether or not to simplify the rate exprs
                    -> DNAMtransition a ApparentRate -- ^ Input transition
                    -> DNAMtransition a DNAMspeed    -- ^ Output transition
translateTransition reduceRates trans
   | Passive == (rateKind $ transSpeed trans) =
      error $ unwords [ "An activity may not have a combined passive rate"
                      , "all activities which passively observe must cooperate"
                      , "over that activity with at least one component which"
                      , "actively performs that activity."
                      ]
   | otherwise                                =
      DNAMtrans { transName       = transName       trans
                , transActionKind = transActionKind trans
                , transConditions = conditions
                , transActions    = transActions    trans
                , transSpeed      = speed
                , transPriority   = transPriority   trans
                }
   where
   conditions = transConditions trans
   speed      = translateSpeed conditions $ transSpeed trans

   translateSpeed :: [ DNAMcondition ] -> ApparentRate -> DNAMspeed
   translateSpeed _ (SingleRate r) = speedOfRate r
   translateSpeed conds r          = DNAMrate $ cExprOfApparentRate conds r
    
   cExprOfApparentRate :: [ DNAMcondition ] -> ApparentRate -> DNAMcexp
   cExprOfApparentRate _     (SingleRate r)                 = cExprOfRate r
   cExprOfApparentRate conds (CompoundRate _ r1 r2 raP raQ)
      | checkForR1 == False =
         error "Mixing passive and active rates"
      | checkForR2 == False =
         error "Mixing passive and active rates"
      | otherwise           =
      Cmult (Cdiv  (cExprOfApparentRate conds r1) raPexp)
            (Cmult (Cdiv (cExprOfApparentRate conds r2) raQexp)
                   minimumFactor
            )
      where
      minimumFactor =
         {-
            So here we use that if r1 is Passive then all the rates
            in raP must be passive. Also if r1 is Active then all the
            rates in raP must be active. And similarly for r2 and raQ.
         -}
         case (rateKind r1, rateKind r2) of
            (Passive, Passive) -> (dnamMinimum raPexp raQexp)
            (Passive, Active)  -> raQexp
            (Active, Passive)  -> raPexp
            (Active, Active)   -> (dnamMinimum raPexp raQexp)
            
      checkForR1 = checkRateAndDivisor r1 raP
      checkForR2 = checkRateAndDivisor r2 raQ

      raPexp = cExpOfRateDivisor raP
      raQexp = cExpOfRateDivisor raQ

      cExpOfRateDivisor :: RateDivisor -> DNAMcexp
      cExpOfRateDivisor []          = 
         -- Since the given process can at least perform this
         -- activity by the given transition.
         error "Should never have zero rate divisors"
      cExpOfRateDivisor rateDivisor =
         foldr1 Cadd $ map makeOne rateDivisor

      -- Makes a c expression out of one of the rate divisors.
      -- Recall a rate divisor is one of the rates at which it is
      -- possible for the given process to do the given activity
      makeOne :: ( [ DNAMcondition ], ApparentRate ) -> DNAMcexp
      makeOne (rConds, r) =
         case filteredConditionals of
            []       -> cExprOfApparentRate conds r
            remConds ->
               Cifte condExp (cExprOfApparentRate conds r) $ Creal 0.0
               where
               condExp  = foldr1 Cand condExps
               condExps = map cExpOfDNAMcond remConds
         where
         -- This is where we use the 'reducedRates' parameter if we are
         -- not to reduce the rates then we do not filter the conditionals
         filteredConditionals
            | reduceRates = filter (not . condInTransitionConds) rConds
            | otherwise   = rConds
 
      condInTransitionConds :: DNAMcondition -> Bool
      condInTransitionConds c = any (equalCond c) conds
   
      equalCond :: DNAMcondition -> DNAMcondition -> Bool
      equalCond (DNAMprocpresent ident1) (DNAMprocpresent ident2) =
         ident1 == ident2
      equalCond _ _                                               =
         False


speedOfRate :: ParsedRate -> DNAMspeed
speedOfRate (RateTop)                = error "No speeds should be infty"
speedOfRate (RateTimed r)            = DNAMrate $ cExprOfRateExp r
speedOfRate (RateImmediate (Just d)) = DNAMweight $ Creal d
speedOfRate (RateImmediate Nothing)  = DNAMweight $ Creal 1.0

cExprOfRate :: ParsedRate -> DNAMcexp
cExprOfRate (RateTop)                = {- Cinfty $-} Creal 1.0
cExprOfRate (RateImmediate (Just x)) = Creal x
cExprOfRate (RateImmediate Nothing ) = Creal 1.0
cExprOfRate (RateTimed r)            = cExprOfRateExp r

{- 
   So RateProc does not currently work here, this is wrong as it should take
   into account all the renamed processes which are the given component.
   That is @P <a> P@ will get turned into @P1 <a> P2@ so the rate expression
   @if P then r1 else r2@ should get turned into
   @if P1 || P2 then r1 else r2@
-}
cExprOfRateExp :: ParsedRateExp -> DNAMcexp
cExprOfRateExp (RateId x _)          = Cident $ translateRateIdentifier x
cExprOfRateExp (RateNum x)           = Creal x
cExprOfRateExp (RateBexp s r1 r2)    =
   (cBinopOfString s) (cExprOfRateExp r1) (cExprOfRateExp r2)
cExprOfRateExp (RateProc _ident)     = 
   error "Functional rate not quite implemented yet, see cExprOfRateExp"
cExprOfRateExp (RateCond _p _r1 _r2) =
   error "Functional rate not quite implemented yet, see cExprOfRateExp"
cExprOfRateExp (RateDisabled)        =   
   -- Maybe just Creal 0.0 ?
   error "Disabled rate in cExprOfRateExp"
-- cExprOfRateExp _            = CFreeForm "unimplemented rate expression"


translateRateIdentifier :: RateIdentifier -> DNAMidentifier
translateRateIdentifier = (flip prefixQName) "PEPA_RATE__"

cBinopOfString :: String -> DNAMcexp -> DNAMcexp -> DNAMcexp
cBinopOfString "+" = Cadd
cBinopOfString "-" = Csub
cBinopOfString "*" = Cmult
cBinopOfString "/" = Cdiv
cBinopOfString _   = error "Undefined binary operator  in cBinopOfString"

{-
   Checks if the given rate and all the rates used in the divisor
   are all passive or all active.
-}
checkRateAndDivisor :: ApparentRate -> RateDivisor -> Bool
checkRateAndDivisor r ra =
   all (== (rateKind r)) $ rateKindsOfDivisor ra
   where
   rateKindsOfDivisor :: RateDivisor -> [ RateKind ]
   rateKindsOfDivisor = map (rateKind . snd)

type IntermediateGPS = GenericParallel ( DNAMgenParTransitions )


{-|
   The main exported function. Converts a pepa model into a
   hydra model.
-}
dnamFromPepa :: ParsedModel -> DNAMmodel ParsedAction ApparentRate
dnamFromPepa pModel@(ParsedModel rDefs _pDefs _mainComp) =
   uniquifyTransitions dnamModel
   where
   dnamModel = DNAMmodel { modelHeaders     = [ DNAMheader header ]
                         , modelStateVector = stateVector
                         , modelConstants   = map translateRateDef rDefs
                         , modelInitial     = initial
                         , modelTransitions = transitions
                         , modelInvariants  = []
                         }
   
   header    =  "/* it's good to be back */"
   genModel :: IntermediateGPS
   genModel  =  toGenericParallel pModel

   DNAMgenParTransitions transitions = convertGPS genModel

   initialConcentations = getInitialConcentrations pModel
   initial              = initialFromConcs     initialConcentations
   stateVector          = stateVectorFromConcs initialConcentations

   initialFromConcs :: ConcentrationMap -> DNAMinitialVector
   initialFromConcs = map makeInitial 

   makeInitial :: (ParsedComponentId, Int) -> DNAMcassignment
   makeInitial (ident, i) = DNAMcassignment ident $ Cconstant i

   stateVectorFromConcs :: ConcentrationMap -> DNAMstateVector
   stateVectorFromConcs = map $ (DNAMstateDesc CShort) . fst

   translateRateDef :: RateSpec -> DNAMconstantDef 
   translateRateDef (rIdent, rExp) =
      DNAMconstantDef (translateRateIdentifier rIdent)
                      (cExprOfRateExp rExp)



{- This function should be generalised and moved into GenericParallel -}
convertGPS :: IntermediateGPS -> DNAMgenParTransitions
convertGPS (SEQ trans)              = trans
convertGPS (PAR left actions right) =  
   composeDNAMtransLists leftTrans rightTrans actions
   where
   leftTrans  = convertGPS left
   rightTrans = convertGPS right
convertGPS (ARRAY left size actions) =
   multiplyDNAMtransLists leftTrans (fromInteger size) actions
   where
   leftTrans = convertGPS left

{-
   The should multiple the transition lists appropriately for a
   number of identical components cooperating over some action set.
   Basically this should implement process arrays.

   Oh, I think actually the number does not matter, because each transition's
   rate will be based on the number that are in that state not the number to
   begin with.
-}
multiplyDNAMtransLists :: DNAMgenParTransitions -> Int -> [ ParsedAction ]
                       -> DNAMgenParTransitions
multiplyDNAMtransLists _ _ =
   error "Not yet implemented, process arrays in compiling Hydra"

{-
   Composes two transition lists, combining any shared transitions.
-}
composeDNAMtransLists :: DNAMgenParTransitions
                      -> DNAMgenParTransitions
                      -> [ ParsedAction ]
                      -> DNAMgenParTransitions
composeDNAMtransLists (DNAMgenParTransitions left)
                      (DNAMgenParTransitions right) actions =
   DNAMgenParTransitions transitions
   where
   transitions = concat [ leftNonCoop
                        , rightNonCoop
                        , concat coopTrans
                        ]
   coopTrans    = map (cooperateActionTransitions leftCoop rightCoop) actions

   leftNonCoop  = snd leftParts
   leftCoop     = fst leftParts
   leftParts    = partition transitionCooperates left

   rightNonCoop = snd rightParts
   rightCoop    = fst rightParts
   rightParts   = partition transitionCooperates right

   transitionCooperates :: DNAMtransition ParsedAction ApparentRate -> Bool
   transitionCooperates t = elem (transActionKind t) actions


cooperateActionTransitions :: [ DNAMtransition ParsedAction ApparentRate ]
                           -> [ DNAMtransition ParsedAction ApparentRate ]
                           -> ParsedAction
                           -> [ DNAMtransition ParsedAction ApparentRate ]
cooperateActionTransitions ts1 ts2 action =
    [ constructDNAMsharedTrans (getRateDivisor t1 ts1 action) 
                                   (getRateDivisor t2 ts2 action) t1 t2 
      | t1 <- findDNAMact action ts1, t2 <- findDNAMact action ts2 ]

{-
   Obviously an entirely incorrect definition so far, but this is
   just a place holder.

getApparentRate :: DNAMtransition ParsedAction ApparentRate
                -> [ DNAMtransition ParsedAction ApparentRate ]
                -> ParsedAction
                -> ParsedRateExp
getApparentRate _ _ _ = RateImmediate Nothing
-}

{-
   calculate apparent rate making sure to exclude any superfluous conditions 
   that will be logically true given the conditions to fire that 
   particular transition. I don't know why we are filtering out those
   conditions which are trivially true?
   In order to actually do that filtering we will need to update the data type
   for 'DNAMtrans' to include a list of conditions rather than just one.
   For now we will just leave it and worry about simplification later.
-}
getRateDivisor :: DNAMtransition ParsedAction ApparentRate
                -> [ DNAMtransition ParsedAction ApparentRate ]
                -> ParsedAction
                -> RateDivisor
getRateDivisor _trans1 transitions action =
   case findDNAMact action transitions of
      []                         -> error nullTrans
      [ one ]                    -> [ ( transConditions one, transSpeed one ) ]
      relevantTrans              -> map getCondsAndApparentRate relevantTrans
   where
   getCondsAndApparentRate :: DNAMtransition ParsedAction ApparentRate
                          -> ( [ DNAMcondition ], ApparentRate )
   getCondsAndApparentRate trans =
      ( transConditions trans, transSpeed trans )
{-
   {-
      Given a list of conditions filter out those that are already
      in the condition list of firing action (that is those that
      are given into this function). We also make sure that the
      list of remaining conditions is not the empty list.
   -}
        mkConditions :: [ DNAMtransCond ] -> [ DNAMtransCond ]
        mkConditions cs = rewrite $ filter (\c -> not $ elem c cs') cs 
              
	rewrite :: [DNAMtransCond] -> [DNAMtransCond]
        rewrite [] = [ DNAMtt ]
        rewrite xs = xs
-} 
   nullTrans  =  "getDNAMapparentRate: no synchronising actions " ++
                 "in cooperating components"

{-
   The old function I got from the ipc source
getDNAMapparentRate :: DNAMtrans DNAMrateExpr -> [DNAMtrans DNAMrateExpr] 
                    -> PEPAact -> DNAMapparentRate
getDNAMapparentRate (DNAMt { transConditions = cs'}) ts a 
    | null relevantTrans          = error nullTrans
    | (length relevantTrans == 1) = DNAMarRt $ getTransFiring 
                                    (head relevantTrans)
    | otherwise                   = 
        DNAMarExpr $ map mkApparentRateExpr $ 
                     map getCondsAndApparentRate relevantTrans
        where
        mkApparentRateExpr :: ([ DNAMtransCond ], DNAMrateExpr)
                           -> (DNAMtransCond, DNAMrateExpr)
        mkApparentRateExpr (condList, ar) =
            ( foldr1 DNAMand $ mkConditions condList, ar )
                              
        getCondsAndApparentRate :: DNAMtrans DNAMrateExpr 
                                -> ( [ DNAMtransCond ], DNAMrateExpr )
        getCondsAndApparentRate trans =
            (transConditions trans, ar)
            where DNAMmarkov ar = transFiring trans

        {-
          Given a list of conditions filter out those that are already
          in the condition list of firing action (that is those that
          are given into this function). We also make sure that the
          list of remaining conditions is not the empty list.
        -}
        mkConditions :: [ DNAMtransCond ] -> [ DNAMtransCond ]
        mkConditions cs = rewrite $ filter (\c -> not $ elem c cs') cs 

  relevantTrans = findDNAMact a ts
              
  rewrite :: [DNAMtransCond] -> [DNAMtransCond]
        rewrite [] = [ DNAMtt ]
        rewrite xs = xs
 
        nullTrans  =  "getDNAMapparentRate: no synchronising actions " ++
                      "in cooperating components"
-}


{-
  Returns from a list of DNAM transitions all those whose action equals
  the given action.
-}
findDNAMact :: ParsedAction
           -> [ DNAMtransition ParsedAction ApparentRate ]
           -> [ DNAMtransition ParsedAction ApparentRate ]
findDNAMact a = filter $ (==a) . transActionKind

-- active/passive and active/active rate calculation
constructDNAMsharedTrans :: RateDivisor
                         -> RateDivisor
                         -> DNAMtransition ParsedAction ApparentRate 
                         -> DNAMtransition ParsedAction ApparentRate 
                         -> DNAMtransition ParsedAction ApparentRate 
constructDNAMsharedTrans rateDiv1 rateDiv2 trans1 trans2
   | a1 /= a2  = error differentActions
   -- | p1 /= p2  = error unequalPriorities
   | otherwise = sharedTrans
   where 
   DNAMtrans { transName       = l1
             , transActionKind = a1
             , transConditions = cs1
             , transActions    = as1
             , transPriority   = p1
             , transSpeed      = speed1 } = trans1
   DNAMtrans { transName       = l2
             , transActionKind = a2
             , transConditions = cs2
             , transActions    = as2
             , transPriority   = p2
             , transSpeed      = speed2 } = trans2
   sharedTrans   = DNAMtrans { transName       = concatQNames l1 "__" l2
                             , transActionKind = a1
                             , transConditions = sharedCondition
                             , transActions    = sharedActions
                             , transPriority   = sharedPriority
                             , transSpeed      = sharedSpeed }
   sharedCondition   = cs1 ++ cs2 -- nub (cs1 ++ cs2)
   sharedActions     = foldr insertDNAMtransAction as1 as2
   sharedKind        = combineRateKinds speed1 speed2
   sharedPriority    = min p1 p2
   sharedSpeed       = 
      case (speed1, speed2) of
         ( SingleRate (RateImmediate w1), SingleRate (RateImmediate w2)) -> 
            SingleRate $ RateImmediate (combineWeights w1 w2)
         ( SingleRate (RateImmediate _), _)                              -> 
            error "attempt to mix instantaenous with timed actions"
         ( _, SingleRate (RateImmediate _))                              ->
            error "attempt to mix instantaenous with timed actions"
         ( _, _ )                                                        ->
            CompoundRate sharedKind speed1 speed2 rateDiv1 rateDiv2

   combineWeights :: Maybe Double -> Maybe Double -> Maybe Double
   combineWeights (Just w1) (Just w2) = Just $ w1 + w2
   combineWeights (Just w1) Nothing   = Just $ w1 + 1.0
   combineWeights Nothing   (Just w2) = Just $ 1.0 + w2
   combineWeights Nothing   Nothing   = Just 2.0

   combineRateKinds :: ApparentRate -> ApparentRate -> RateKind
   combineRateKinds apRate1 apRate2 =
      case (rateKind apRate1, rateKind apRate2) of
         (Passive, Passive) -> Passive
         (_, _)             -> Active
   -- should of course be something like
   {- case (speed1, speed2) of
         (DNAMrate r1, DNAMrate r2)     ->
            DNAMrate $ mkDNAMrateExpr (r1, ar1) (r2, ar2) 
         (DNAMweight w1, DNAMweight w2) ->
            DNAMweight $ w1 + w2
         (_, _)                         ->
            error "Mixing instantaneous and timed actions"
   -}
   differentActions  = "IPC.constructDNAMsharedTrans:" ++
                       " not joining equal PEPA action-type transitions"
   -- unequalPriorities = "IPC.constructDNAMsharedTrans: unequal priorities"


{-
   This function in the original ipc compiler attempted to insert a
   dnam action into a list of such actions by checking whether or not
   the action already existed and if so combining them.
   So for example if we had:
   @ P = P + 1 @
   and
   @ P = P + 1 @
   then instead of returning the list with both actions in it we would
   combine it into one action as such:
   @ P = P + 2 @
   But I don't really see the point in this so I'm just going to use
   concatenation.
   
   Here is the original function
insertDNAMtransAction :: DNAMtransAction
                      -> [DNAMtransAction] 
                      -> [DNAMtransAction]
insertDNAMtransAction t@(DNAMta lbl (lbl1, i)) ts
    | isNothing fnd = (t:ts)
    | (lbl1 /= lbl2) || (lbl1 /= lbl) = error invalidAction
    | (i+j) == 0    = ts' -- uncomment to remove {P := P + 0; } style actions
    | otherwise = (DNAMta lbl (lbl, i+j):ts') where
      fnd = selectTerm (t ==) ts
        (DNAMta _ (lbl2, j), ts') = fromJust fnd
   where
   invalidAction =  "IPC.insertDNAMtransAction: " ++
                    "mismatched configuration of transition actions. " ++
                    "ie, p1 := p2 + k for p1 != p2"
-}
insertDNAMtransAction :: DNAMaction -> [ DNAMaction ] -> [ DNAMaction ]
insertDNAMtransAction = (:)


{-
   Note here the model must be in a normal form such that each successor
   of a prefix component is an identifier component.
   The successor map should only include those successors whose source
   are reachable from the given component.
-}
fromPepaComponent :: SuccessorMap 
                  -> ParsedComponentId
                  -> DNAMgenParTransitions
fromPepaComponent sMap _ident =
   DNAMgenParTransitions $ map translatePepaTransition sMap


{- 
   Translate a Pepa transition into a DNAMtransition
-}
translatePepaTransition  :: Successor
                         -> DNAMtransition ParsedAction ApparentRate
translatePepaTransition (identSource, (trans, identTarget)) =
   DNAMtrans { transName       = transitionName
             , transActionKind = aKind
             , transConditions = [ condition ]
             , transActions    = commands
             , transSpeed      = SingleRate $ pepaTransRate trans
             , transPriority   = pepaTransPriority trans
             }
   where
   transitionName = concatQNames identSource middleBit identTarget
   middleBit      = "__" -- ++ (hprint $ nameOfAction aKind) ++ "_"
   aKind          = pepaTransAction trans
   condition      = DNAMprocpresent identSource

   commands
      | identSource == identTarget = []
      | otherwise                  = [ DNAMidentDecr identSource
                                     , DNAMidentIncr identTarget
                                     ]

   
{-|
   Now this is the part of the compilation that is at best a little
   suspect. I would definitely like to think up a better way to do
   this, but for now this will suffice.
   Basically we have added in probe components and these have now been
   translated. However the names of them will have been qualified.
   So essentially we take in the name (for the time being we allow
   just the one measurement specification) of and search the list
   of components (defined in the Hydra model) and use these to create
   a measurement specification.
   
   Note that the type of this is likely to change to reflect the
   possibility of adding in more than just a single passage-time
   measurement.

-}
addMeasurementSpecs :: DNAMsolutionControl
                    -> [ DNAMperformMeasure ]
                    -> (DNAMmodel ParsedAction b)
                    -> (DNAMmodelFile ParsedAction b)
addMeasurementSpecs solutionControl measures model =
   DNAMmodelFile model solutionControl newMeasures
   where
   newMeasures  = map makeNewMeasure measures

   makeNewMeasure :: DNAMperformMeasure -> DNAMperformMeasure
   makeNewMeasure (DNAMpassage pt)   =
      DNAMpassage $ pt { passageSourceCond = newSourceCond
                       , passageTargetCond = newTargetCond
                       }
      where
      newSourceCond = renameCondition $ passageSourceCond pt
      newTargetCond = renameCondition $ passageTargetCond pt
   makeNewMeasure (DNAMtransient tm) =
      DNAMtransient $ tm { transientSourceCond = newSourceCond
                         , transientTargetCond = newTargetCond
                         }
      where
      newSourceCond = renameCondition $ transientSourceCond tm
      newTargetCond = renameCondition $ transientTargetCond tm
   makeNewMeasure (DNAMsteady sm)    =
      DNAMsteady $ sm { steadyCondition = newSteadyCond }
      where
      newSteadyCond = renameCondition $ steadyCondition sm
   makeNewMeasure (DNAMcount cm)     =
      DNAMcount $ cm { countTransitions = makeCountNames trans }
      where
      trans = countTransitions cm
                                     

   componentNames = map nameOfStateDesc $ modelStateVector model
   nameOfStateDesc :: DNAMstateDesc -> DNAMidentifier
   nameOfStateDesc (DNAMstateDesc _ ident) = ident

   renameCondition :: DNAMcondition -> DNAMcondition
   renameCondition (DNAMprocpresent ident) =   
      case find (\s -> (showOrig s) == (showOrig ident)) componentNames of
         Just s  -> DNAMprocpresent s
         Nothing -> error noComponentErr
   renameCondition _                       =
      error "Compile.Hydra.addMeasurementSpecs: invalid condition"
   

   noComponentErr = "Compile.Hydra.addMeasurementSpecs: missing component name"

   makeCountNames :: [ QualifiedName ] -> [ QualifiedName ]
   makeCountNames names =
      map transName $ filter isCorrectAction (modelTransitions model)
      where
      isCorrectAction :: DNAMtransition ParsedAction b -> Bool
      isCorrectAction t = elem (nameOfAction $ transActionKind t) names