{-|
   A module for converting Pepa models to prism models.
-}

module Language.Pepa.Compile.Prism
   ( prismFromPepa
   , makePrismFile
   )
where

{- External Library Imports -}
{- Standard Library Imports -}
{- Local Module Imports -}
import Language.Pepa.Syntax
   ( QualifiedName   ( .. )
   , ParsedModel
   , ParsedAction
   )
import Language.Prism.Syntax
   ( PrismModelFile  ( .. )
   , PrismModel      ( .. )
   , PrismModelKind  ( .. )
   , PrismModule     ( .. )
   , PrismVariable   ( .. )

   , PrismTransition ( .. )
   , PrismAction
   , PrismCondition
   , PrismRate
   )
import Language.Hydra.Syntax
   ( DNAMmodel             ( .. )
   , DNAMinitialAssign
   , DNAMcassignment       ( .. )
   , DNAMcexp              ( .. )
   , DNAMtransition        ( .. )
   , DNAMaction
   , DNAMcondition         ( .. )
   )
import Language.Hydra.Transform
   ( translateRates )

import Language.Pepa.Compile.Hydra
   ( dnamFromPepa
   , ApparentRate
   , translateTransition
   )

{- End of Imports -}

{-| The main function exported from this module. Translates a pepa
   model into a prism model.
   The first option is whether or not to reduce the rate expressions.
-}
prismFromPepa :: Bool -> ParsedModel -> PrismModel
prismFromPepa reduceRates = 
   prismFromDnam . -- Translate from dnam to prism
   (translateRates transTrans) . -- Translate the rates in the dnam model
   dnamFromPepa -- Translate from pepa to dnam
   where
   transTrans      = translateTransition reduceRates
   

{-| Turns a prism model in to a full-blown prism file ready for output.
   Note that the type of this is likely to change.
-}
makePrismFile :: PrismModel -> PrismModelFile
makePrismFile = PrismModelFile

{-
   The translation from hydra models into prism models.
-}
prismFromDnam :: DNAMmodel ParsedAction PrismRate -> PrismModel
prismFromDnam dnamModel = 
   PrismModel PrismCtmc [ mainModule ]
   where
   mainName    = Unqualified "main"
   mainModule  = PrismModule mainName variables transitions
   variables   = map convertVariable $ modelInitial dnamModel
   transitions = map convertTransition $ modelTransitions dnamModel

   {-
      The variables are not quite translated well, certainly not to do
      aggregation. Effectively at the moment we are assuming that all
      variables can take simply 0 or 1 when in fact if aggregation has
      been used, this is not true.
   -}
   convertVariable :: DNAMinitialAssign -> PrismVariable
   convertVariable (DNAMcassignment ident (Cconstant i))
      | i == 0 || i == 1 = 
         PrismVariable { prismVarname = ident
                       , prismVarLow  = 0
                       , prismVarHigh = 1
                       , prismVarInit = i
                       }
      | otherwise        =
         error "Unimplemented aggregation / non 0-1 processes in prism"

   convertVariable (DNAMcassignment _ _) =
      error "A dnam initial assignment should be a constant value"
   convertVariable (DNAMidentIncr _) =
      error "A dnam initial assignment should not be an increment"
   convertVariable (DNAMidentDecr _) =
      error "A dnam initial assignment should not be a decrement"
      

{-
DNAMmodel { modelHeaders     :: [ DNAMheader ]
                               , modelStateVector :: DNAMstateVector
                               , modelConstants   :: [ DNAMconstantDef ]
                               , modelInitial     :: DNAMinitialVector
                               , modelTransitions :: DNAMmodelTransitions a b
                               , modelInvariants  :: [ DNAMmodelInvariant ]
                               }
-}

convertTransition :: DNAMtransition ParsedAction PrismRate
                    ->    PrismTransition
convertTransition trans =
   PrismTransition { prismTransGuard   = conds
                   , prismTransActions = actions
                   , prismTransRate    = transSpeed trans
                   }
   where
   conds   = map convertGuard $ transConditions trans
   actions = map convertAction $ transActions trans

convertGuard :: DNAMcondition -> PrismCondition
convertGuard = id

convertAction :: DNAMaction -> PrismAction
convertAction = id
