{-|
  This module will define the main monad that we will use to pass around
  data within the imperial pepa compiler.
  The main purpose here is to define a monad that will carry around the
  logged information, warnings and errors, in this way most of the major
  functions that each module exports can refer to this monad which should
  then be able to modify with very little effect on the modules which
  use this.

   This module could do with a wee bit of a tidy-up.
-}

module Language.Pepa.IpcControl
   ( MainControl
     -- * Logging types
   , LogRes
   , LogInfo
   , Warning
   , IpcError

     -- * Functions for Main Control.
   , analyseForLog
   
   
   , valueResult
   
   , valueResultWith
   , valuePhase
   , valuePhaseWith
   , resultWarning
   , maybeWarning
   
   , addPhase
   
   , addHeader
   
   , closeResult
   
   , closeJustResult
   , closeResultHeader
   , phaseMapBy
   
   , phaseMap
   , moduleExecuted
   
   , isResult
   
   
   
   , resultError
   , showMainControlA
   , displayMainControlA
   , showMainControlAWith
   , displayMainControlAWith
   , inspectMainControl
   , showMainControlWarnings
   , mainControlNumberWarnings
   
   , closeOrError

   , eitherToMainControl
   )
where

{- External Library Module Imports -}
{- Standard Library Module Imports -}
{- Local Module Imports -}
{- End of Module Imports -}

{-|
  Main control is the type which is passed around between the various
  functions of the compiler. It contains both the side information together
  with the actual result.

  [@todo@]Fatal error should also contain the logging information so that
  we can write a log on an error which might help debug the error.
  Actually it should also contain the warnings as well for the same reason.
  Functions like /mapPhase/ could then add to the logging and possibly
  the warnings the fact that you have made an attempt to get the result
  out and failed.

  [@todo@]So currently in the case of an error we get no log information,
  obviously that sucks because when we get a fatal error that is surely
  when we most require the logging information, so I say that we shoul
  instead include the logging information in the 'FatalError'.

  [@todo@]Also there should be levels of logging, so that for example in condor
  mode, we can log some important information, such as say the rate values
  used, but not all the other stuff which may slow down the computation,
  for example the analysis of all the kinds of model that we have.
-}
data MainControl a = 
    FatalError IpcError LogInfo
  | MainResult a [ Warning ] LogInfo

{-|
  The Logging information is basically just a string, but we could
  actually make it a bit more high level.
  Ultimately I think the log information should be two strings one
  representing the current module and the other representing the
  current log, when we add to it, we should give the current module
  if it is the same as the one stored we just add the string
  otherwise we can put in a little header to start the logging information
  for that module.
-}
type LogInfo = String

{-|
   And let's just make a type class for that
-}
class LogRes a where
   analyseForLog :: a -> LogInfo

{-|
  For now a warning is also just a string but this I would definitely
  like to have as a data type
-}
type Warning = String

{-|
  And the same goes for errors, I would definitely like to have this
  as a data type.
-}
type IpcError = String

{-
  Now we must have a function for placing a value within a "result"
  this is essentially the return method of a monad.
-}
valueResult :: LogRes a => a -> MainControl a
valueResult a = MainResult a [] $ analyseForLog a

{-
  'valueResultWith' is the same as 'valueResult' except instead
  of using the implied 'analyseForLog' function to analyse the
  value we allow the use of a given analysing function.
-}
valueResultWith :: (a -> String) -> a -> MainControl a
valueResultWith anaylse a = MainResult a [] $ anaylse a

{-
  The function 'valuePhase' is similar to 'valueResult' except
  that we give a string describing 
-}
valuePhase :: LogRes a => String -> a -> MainControl a
valuePhase modName a = MainResult a [] logInfo
                       where logInfo = header ++ (analyseForLog a)
			     header  = moduleExecuted modName

{-
  Finally in the section the function 'valuePhaseWith' is analagous
  to 'valuePhase' in the same way that 'valueResultWith' is
  analagous to 'valueResult'.
-}
valuePhaseWith :: (a -> String) -> String -> a -> MainControl a
valuePhaseWith analyse modName a =
    MainResult a [] logInfo
    where logInfo = header ++ (analyse a)
	  header  = moduleExecuted modName

{-
  The function 'resultWarning' returns a result
  but with a warning.
-}
resultWarning :: a -> [ Warning ] -> MainControl a
resultWarning a warns = MainResult a warns ""

{-
  The function 'maybeWarning' returns a result with a
  warning if the warning is there and acts just like 'valueResult'
  if not.
-}
maybeWarning :: a -> Maybe Warning -> MainControl a
maybeWarning a Nothing  = MainResult a [] ""
maybeWarning a (Just w) = MainResult a [ w ] ""

{-
  Failure, call this function to report an error.
-}
resultError :: IpcError -> MainControl a
resultError err = FatalError err ""



{-
  The function 'addPhase' is basically the same as the bind function
  of a monad, except that it allows the beginning of a new phase
  in the compiler.
  The arguments consist of : The old result, the module string,
  the new result and the new logging information.
-}
addPhase :: MainControl a -> String -> b -> String -> MainControl b
addPhase (FatalError err logs) _ _ _                         = 
    FatalError err logs
addPhase (MainResult _ warns logInfo) modName res newLogInfo =
    MainResult res warns $ logInfo ++ header ++ newLogInfo
    where header = moduleExecuted modName 


addHeader :: String -> MainControl a -> MainControl a
addHeader _ mc@(FatalError _ _)                 = mc
addHeader newLog (MainResult res warns logInfo) =
    MainResult res warns $ newLog ++ logInfo

{-
  Also a function to return the value within the "result". 
  It returns a pair of both the value held within the result and the
  side information. This is essentially how we "get out of" our monad.
  This must be used with caution, if this is called then unless you 
  explicitly take care of the side information yourself it will be lost.
-}
closeResult :: MainControl a -> (a, [Warning], LogInfo)
closeResult (FatalError err _)           = error err
closeResult (MainResult a warnings logS) = (a, warnings, logS)

closeOrError :: MainControl a -> Either IpcError (a, [Warning], LogInfo)
closeOrError (FatalError err _)           = Left err
closeOrError (MainResult a warnings logS) = Right (a, warnings, logS)

{-
  returns just the result without the log information or any warnings.
  Use this carefully, should only be used if the 'MainControl' is
  being passed on elsewhere
-}
closeJustResult :: MainControl a -> a
closeJustResult (FatalError err _)   = error err
closeJustResult (MainResult a _ _)   = a

{-
  The same as 'closeResult' except that it allows the addition
  of a standard header.
-}
closeResultHeader :: String -> MainControl a -> (a, [Warning], LogInfo)
closeResultHeader s res = 
    (a, warns, s ++ logInfo)
    where (a, warns, logInfo) = closeResult res

{-
  This section contains functions that allow the mapping of functions
  that operate on ordinary values to operating on values contained
  within a 'MainControl'.

  The function 'phaseMap' essentially takes in a function from
  'a' to 'b' and turns it into a function from a 'MainControl a'
  to a 'MainControl b', the preceding arguments are there to allow
  the adding of log information the first is a function from the
  result type of the important function argument to a string and the
  second is a string which should be the module name.
-}
phaseMapBy :: (b -> String) -> String -> (a -> b) 
	    -> MainControl a -> MainControl b
phaseMapBy _ _ _ (FatalError err logs)                = FatalError err logs
phaseMapBy analyseB modName f (MainResult a warns lg) = 
    MainResult b warns newLog
    where b      = f a
	  newLog = lg ++ (moduleExecuted modName) ++ (analyseB b)

{-
  The same as 'phaseMapBy' except that we use the log result's
  analyse function.
-}
phaseMap :: LogRes b => String -> (a -> b) 
	 -> MainControl a -> MainControl b
phaseMap = phaseMapBy analyseForLog

{-
  We now make the main control type an instance of the monad class.
  This is purely to allow the use of the do notation to string together
  calls to functions that return the main result.
-}

bindControl :: MainControl a -> (a -> MainControl b) -> MainControl b
bindControl (FatalError err logs) _     = FatalError err logs
bindControl (MainResult a warns logs) f =
    case f a of
    (FatalError err logs')       -> FatalError err $ logs ++ logs'
    (MainResult b warns' logs')  -> 
	MainResult b (warns ++ warns') (logs ++ "\n" ++ logs')

instance Monad MainControl where
    (>>=)    = bindControl
    return a = MainResult a [] ""


{-
  @todo{Write a brief tutorial on how to use 'IpcControl.MainControl'}
-}

{-
  The following are for printing the values held within a 'MainControl'.
  This means that we do not have to come outside of the 'MainControl'
  monad to print, for example the parsed model, to the screen.
-}
showMainControlA :: Show a => MainControl a -> String
showMainControlA (MainResult a _ _)   = show a
showMainControlA (FatalError err _)   = err

displayMainControlA :: Show a => MainControl a -> IO ()
displayMainControlA (MainResult a _ _) = putStrLn $ show a
displayMainControlA (FatalError err _) = putStrLn err

showMainControlAWith :: (a -> String) -> MainControl a -> String
showMainControlAWith f (MainResult a _ _)   = f a
showMainControlAWith _ (FatalError err _)   = err

displayMainControlAWith :: (a -> String) -> MainControl a -> IO ()
displayMainControlAWith f (MainResult a _ _) = putStrLn $ f a
displayMainControlAWith _ (FatalError err _) = putStrLn err

{-
  Inspect the value within a main control, the caller provides a default
  value should the Main Control contain no value, that is if it is
  an error.
-}
inspectMainControl :: a -> (b -> a) -> MainControl b -> a
inspectMainControl def _ (FatalError _ _) = def
inspectMainControl _ f (MainResult a _ _) = f a

{-
  \subsection{Auxiliary functions}
  A few auxiliary functions used throughout this module.
-}
moduleExecuted :: String -> String
moduleExecuted modName = "Module: " ++ modName ++ " - executed\n\n"


{-
  Inspection to see if we have a result to pass or or an error.
-}
isResult :: MainControl a -> Bool
isResult (FatalError _ _)   = False
isResult (MainResult _ _ _) = True

{-
  Show the current set of warnings.
-}
showMainControlWarnings :: MainControl a -> String
showMainControlWarnings (FatalError _ _)       = 
    "There is currently no result, "++
    "this means there a fatal error with your model"
showMainControlWarnings (MainResult _ warns _) =
    unlines warns

{-
  Return the number of warnings gathered so far
-}
mainControlNumberWarnings :: MainControl a -> Int
mainControlNumberWarnings (FatalError _ _)       = 0
mainControlNumberWarnings (MainResult _ warns _) = length warns



{-
  @todo{Read this and implement it for logging}
  A log should not be a string but a list of log elements.
  A log element is defined by a data type something like:
  '
  data LogElement = Misc Int String
                  | ProcessConcentrations String
                  | CommandLineOptions String
  '
  Then for each logging event either it is explicitly named
  in the 'LogElement' type or use 'Misc x' where 'x'
  is the "level" of logging.
  Then a command line options can set the level of logging, only
  those elements with a higher level of logging will be output
  within the log file.
  For the explicitly named elements these have a default logging
  level which can be overridden by a command line option.
-}

eitherToMainControl :: Show err => Either err a -> MainControl a
eitherToMainControl (Right a)  = MainResult a [] []
eitherToMainControl (Left err) = FatalError (show err) ""