{-|
  This is the new parser module built using the [-parsec-] parser
  cominator library.
-}
module Language.Pepa.Parser
   ( parseRateArrayOption
   , parseRateOption
   , parseActionArrayOption
   , parseCommaSeparatedLowers
   , parseRateExprOption


   , blindRun
   , controlParseResult
   , parsePepaFile
   --, parsePepaWith
   --, parseProbe
   
   -- * Individual parser definitions
   , fileParser
   , pepaFile
   , pepaComponent
   , pepaBeginProcessDef
   , pepaComponentDef
   , pepaActionId
   , pepaComponentId
   , pepaRateId
   , pepaRateDefId
   , pepaRate
   , pepaRateExp
   
   -- * Utility functions or parsers
   , maybeOption
   , parseFile

   -- * Identifier parsers
   , allUpperId
   
   -- * The token parser parsers which use the lexer specific to pepa
   , intParser
   , symbol
   , comma
   , braces
   )
where

{- Standard imported libraries -}
import System.Directory
   ( doesFileExist )
import Text.ParserCombinators.Parsec
   ( Parser
   , CharParser
   , GenParser
   , ParseError
   , parse
   
   , ( <|> )
   , ( <?> )
   , try
   
   , eof
   , string
   , alphaNum
   , char
   , upper
   , lower
   
   , optional
   , many
   , many1
   , sepBy
   , sepBy1
   )
import qualified Text.ParserCombinators.Parsec.Token as StdToken
import Text.ParserCombinators.Parsec.Expr
    ( buildExpressionParser
    , Assoc (..)
    , Operator (..)
    )
import Text.ParserCombinators.Parsec.Language
    ( haskellStyle )

{- Local imported libraries -}
import Language.Pepa.Utils
    ( mapFst
    , mapSnd
    )
import qualified Language.Pepa.PrivateTokenParser as PrvToken

import Language.Pepa.Syntax 
   ( RateIdentifier
   , ActionIdentifier
   , QualifiedName     ( .. )
   , ParsedModel       ( .. )
   , RateSpec
   , ProcessDef
   , ParsedDefinition  ( .. )
   , ParsedComposition
   , ParsedComponent   ( .. )
   , BeCondition       ( .. )
   , ParsedTrans       ( .. )
   , ParsedPriority
   , defaultPepaPriority
   , ParsedComponentId
   , ParsedRate        ( .. )
   , ParsedRateExp     ( .. )
   , ParsedCond        ( .. )
   , ParsedAction      ( .. )
   , nameOfAction
   )
import Language.Pepa.IpcControl
   ( MainControl
   , resultError
   )
{- End of imports -}


{-|
   Parses a file by first checking if the file exists and then
   reading in it's contents and finally parsing the contents
   with the name of the file together with the contents.

   We also return a string as the error, since this allows us
   to return a (string) error in the case that the file does
   not exit.
-}
parseFile :: Parser a -> FilePath -> IO (MainControl a)
parseFile parser file =
   do fileExists <- doesFileExist file
      if fileExists
         then do contents <- readFile file
                 let pResult = parse parser file contents
                 return $ controlParseResult pResult
         else return (resultError $ "file: " ++ file ++ " does not exist")
      



{-
  Blindly runs the given parser expecting it to succeed.
  The first argument is the error string to give to the failure
  function 'error' in the event that the parser fails. 
  This function is most useful for simple parsers within the program
  such as the parsing of command line options.
  Note that the actual parse error returned from the parser itself is
  ignored, hence the user will not recieve a potentially useful message
  such as 'expecting open bracket'.
-}
blindRun :: String -> Parser a -> String -> a
blindRun err p input = case (parse p "" input) of
                       Left err2   -> error (err ++ ": " ++  show err2)
                       Right x     -> x

{-
  A function to turn a parser into a parser that will accept a file
  as input. This function must be careful to skip over white space at
  the beginning, the given parser should take care of the whitespace
  for later, it will normally do this using the 'Parsec.Token' library
  to build its parsers. See the documentation for the parsec library
  for further details.
-}
fileParser :: Parser a -> Parser a
fileParser parser = do whiteSpace
                       x <- parser
                       eof
                       return x

{-
  Similarly we can parse a file and return a main result.
runParseResult :: LogRes a => Parser a -> FilePath -> String 
	       -> Ipc.MainControl a
runParseResult p fileName input =
    case parse p fileName input of
    Left err    -> Ipc.resultError $ show err
    Right value -> Ipc.valueResult value
-}



{-
  We now begin with the utility parsers that are used out with the
  main parser for pepa files. These include for example the parsing
  of command line options.
  Parses the rate command line option argument, which is of the form
  @ r=0.5,1.5,6.9 @
  the integer given as the first argument is used to select which
  value of the comma separated list of values is returned.
-}

{-|
  The action list parser will produce a list of parsed actions but
  these are not just identifiers, so we have to map them into identifiers.
  We probably should store them in the command line options as parsed actions.
-}
parseActionArrayOption :: String -> [ ActionIdentifier ]
parseActionArrayOption input =
   map nameOfAction actions
   where
   err     = "error parsing a pepa action list option"
   actions = blindRun err pepaActionList input


{-|
   Parses in a list of comma separated names which all begin
   with a lower case letter.
   The first argument is the error to give the second
   the input.
-}
parseCommaSeparatedLowers :: String -> String -> [ String ]
parseCommaSeparatedLowers err =
   blindRun err $ sepBy lowerId comma



{-|
  In contrast to that below, when parsing a rate option
  (as opposed to a rate array option) there is only one
  double to return rather than a list of double from which to 
  choose using the process identifier.
 -} 
parseRateOption :: String -> (RateIdentifier, Double)
parseRateOption input =
    blindRun err rateOptionParser input
    where err = "error parsing a rate constant option"


rateOptionParser :: Parser (RateIdentifier, Double)
rateOptionParser = do ident  <- pepaRateId
		      char '='
		      number <- floatNumber
		      return (ident, number)

{-|
  Allows the parsing of a rate option where the rate is set to a
  rate expression rather than a simple constant.
-}
parseRateExprOption :: String -> (RateIdentifier, ParsedRateExp)
parseRateExprOption input =
    blindRun err rateExprOptionParser input
    where err = "error parsing a rate expression option"


rateExprOptionParser :: Parser (RateIdentifier, ParsedRateExp)
rateExprOptionParser = do ident <- pepaRateId
			  char '='
			  expr  <- pepaRateExp
			  eof
			  return (ident, expr)

{-|
   Parse a rate array option, such as
   @ --rates r=0.1,0.2,0.3,0.4,0.5 @
-}
parseRateArrayOption :: String -> (RateIdentifier, [Double])
parseRateArrayOption input = 
    blindRun err rateArrayOptionParser input
    where err = "error parsing a rate constant array option"


rateArrayOptionParser :: Parser (RateIdentifier, [Double])
rateArrayOptionParser = do ident   <- pepaRateId
			   char '='
			   numbers <- rateNumberList
			   return (ident, numbers)

{-
  The rate number list can either be a list of floating point numbers
  separated by commas, or a stepped number list.
  So it looks like either 
  @ r=0.2,0.4,0.6,0.8 @
  or
  @ r=0.2..0.8:0.2 @
  So to parse this we have to parse the first number, since in either case
  we are expecting a number. After this we can decide, if we see a comma then
  a comma list is following if we see a dot then a stepped list is following.
  Note that the first number is passed in to create the remainder parsers and
  that is why we needn't return it in the return expression because it will
  already be incorporated into the results of the two parsers.
-}
rateNumberList :: Parser [Double]
rateNumberList = do first      <- floatNumber
		    remainder  <- ( steppedNumberList first) <|> 
				  ( commaFloatNumberList first )
		    return remainder

{-
  There is a known bug involved with this, it comes from how Haskell
  does double addition. It comes from the representation of doubles
  in particular that of @0.2@. The upshot is that something like
  @0.1 + 0.2@ produces @0.3000000000000004@, which obviously looks
  very bad as a result. I will fix this using a floor function to take
  the number to the nearest say six decimal places.
-}
steppedNumberList :: Double -> Parser [Double]
steppedNumberList start = 
    do _     <- char '.'
       _     <- char '.'
       end   <- floatNumber
       _     <- char ':'
       step  <- floatNumber
       return (computeNumberList start end step)
    where computeNumberList :: Double -> Double 
	          	       -> Double -> [Double]
	  computeNumberList current end step
		 | current > end = []
		 | otherwise     = current : rest
		 where rest = computeNumberList (current + step) end step


{-|
   Changes a parse function such that it will return an
   'IpcControl.MainResult' instead of an
   @Either ParseError a@
-}
controlParseResult :: Either ParseError a -> MainControl a
controlParseResult (Left err) = resultError $ show err
controlParseResult (Right a)  = return a



{-|
    Runs the main pepa model parser over the given input.
    The first argument is the name of the file containing the
    pepa model such that error messages maybe say which pepa
    file contains the error. 
-} 
parsePepaFile :: FilePath                      -- ^ The file to be parsed
              -> String                        -- ^ The contents of the file
              -> Either ParseError ParsedModel -- ^ The result of parsing
parsePepaFile = parse pepaFile


{-|
  The main parser for a pepa model.
-}
pepaFile :: Parser ParsedModel
pepaFile = fileParser pepaModel

{- (remember when this function is exposed to make it a haddock comment)     
  A function to parse a pepa file and return an @IpcControl.MainControl@
  as the result.
parsePepaWith :: FilePath -> String -> Ipc.MainControl ParsedModel
parsePepaWith = runParseResult pepaFile
-}


{-
    A parser for a whole pepa model.
-}
pepaModel :: Parser ParsedModel
pepaModel = 
    do (rateDefs, procDefs) <- pepaDefinitionList
       composition <- pepaComponent
       return $ ParsedModel rateDefs procDefs composition

pepaDefinitionList :: Parser ( [ RateSpec ], [ ProcessDef ] )
pepaDefinitionList = 
   do definitions <- many1 pepaDefinition
      return $ splitDefinitions definitions
   where
   -- I would sort of quite like to give a better error message here.
   -- Actually just somehow fail more gracefully, I think there is
   -- some way in which I can actually return a parser error?
   splitDefinitions :: [ ParsedDefinition a ] -> ([ RateSpec ], [ ProcessDef ])
   splitDefinitions []                       = ([], [])
   splitDefinitions ((RateDef rd) : rest)    = 
      mapFst (rd : ) $ splitDefinitions rest
   splitDefinitions ((ProcessDef pd) : rest) =
      mapSnd (pd : ) $ splitDefinitions rest
   splitDefinitions ((OtherDef _) : _rest)   =
      error "Another kind of definition within a pepa model"


{-
  A pepa model definition, basically can be either a rate definition,
  a process concentration definition or the definition of a 
  sequential process. Currently though we parse process concentration
  definition as a rate definition because they are indistinguishable
  from the point of view of the parser and sort it out in the semantic
  analysis. But see below that I think it would be nice to separate
  them out syntactically.
-}
pepaDefinition :: Parser (ParsedDefinition ())
pepaDefinition = do def <- pepaRateDef <|> pepaComponentDef
		    symbol ";"
		    return def


{-
  Process concentration definitions are parsed as rate definitions
  and we sort it out after parsing, the reason is that it is a bit
  awkward to distinguish between the two, importantly because
  the identifiers are of the same style. It may be worth considering
  changing this actually. We could for example have a syntax something
  like [-concentration p = 0-], unfortunately concentration seems like
  a rather long key word, perhaps instead [-rate r = 0.1-].
-}
pepaRateDef :: Parser (ParsedDefinition a)
pepaRateDef = do ident <- pepaRateDefId
		 symbol "="
		 expr  <- pepaRateExp
		 return $ RateDef (ident, expr)

{-
  A parser for a component definition, so basically a component
  identifier set equal to a *sequential* component expression.

  Note the slightly questionable use of the [-try-] parser
  combinator here. The purpose is that the main composition at
  the end may not be surrounded by parentheses. It may for example
  be [-P <> P-], if it is, and we didn't use [-try-] to parse the
  identifier and the [-=-] sign then *this* parser would successfully
  parse the [-P-] identifier and then fail because there is no [-=-]
  sign. This would cause the [-pepaModel-] parser to come out of the
  [-many1 pepaDefinition-] and attempt to parse the main composition,
  *however* the input [-P-] has already been 'eaten' by this parser.
  Using [-try-] here means that if we fail because the [-=-] sign is
  not present then we will put the [-P-] (or whatever the component
  identifier is) back on to the input and allow the main composition
  to be parsed.
-}
pepaComponentDef :: Parser (ParsedDefinition a)
pepaComponentDef = 
   do ident     <- pepaBeginProcessDef
      component <- pepaComponent
      return $ ProcessDef (ident, component)

pepaBeginProcessDef :: Parser ParsedComponentId
pepaBeginProcessDef = 
   try $ do optional $ symbol "#"
            ident <- pepaProcessDefId
            symbol "="
            return ident


{-
   Parser for components.
   I'm now not so sure that 'buildExpressionParser' is the way to go
   here because of the slight difficulty of parsing a prefix
   component. Also doing this means that we allow the parsing of
   for example @ (a,r).(P <> P) @ which is obvious nonsense.
   I think the better way to go is build up the parser myself,
   for now though this isn't working too badly.
   Also I tried putting in a 'Prefix' line in the @compTable@,
   in order to parse prefix components rather than having them
   as a @compFactor@ but I couldn't get that to work.
-}
pepaComponent :: Parser ParsedComposition
pepaComponent =
    buildExpressionParser compTable compFactor
    <?> "Pepa component"
    where 
    compTable     = [ [ Infix compSum AssocLeft ]
                    , [ Postfix hideTail ]
                    , [ Infix (twoBars <|> angledActions) AssocLeft ]
                    ]

    -- Some pepa tools force @P || P@ to mean cooperation
    -- over the empty set of actions, so we allow this
    -- in addition to @P <> P@ just so that a model does not
    -- need to be written twice.
    twoBars :: Parser (ParsedComponent -> ParsedComponent -> ParsedComponent)
    twoBars = do string "||"
                 return $ makeCoop []
    angledActions :: Parser (ParsedComponent -> ParsedComponent 
                          -> ParsedComponent)
    angledActions = do actions <- angles pepaActionList
                       return $ makeCoop actions

    makeCoop :: [ParsedAction] -> ParsedComponent 
             -> ParsedComponent -> ParsedComponent
    makeCoop actions left right = Cooperation left actions right
             
    compSum :: Parser (ParsedComponent -> ParsedComponent -> ParsedComponent)
    compSum = do symbol "+"
                 return ComponentSum
      
    hideTail = do symbol "/"
		  actions <- braces pepaActionList
		  return $ makeHide actions

    makeHide :: [ParsedAction] -> ParsedComponent -> ParsedComponent
    makeHide actions left = Hiding left actions 

    compFactor  = (try pepaConditional)
                  <|> (try pepaPrefix)
		  <|> processName
                  <|> parens pepaComponent
                  <?> "simple process composition"

    -- Currently we do not allow the following
    -- @ ((a,r).P)[3] @
    -- that is a process array can only be a single identifier
    -- I may relax this and if so the checking here is unnecessary.
    processName = do name      <- pepaComponentId
                     arrayTail <- maybeOption processTail
                     return $ addProcessTail name arrayTail

    -- Okay I'm not completely sure about having 'pepaComponent' as the
    -- second alternative to a 'pepaPrefix'.
    -- As you can see I previously had just a simple
    -- ident component there, but I would quite like to allow things such as
    -- @ (r, 1.0).(P + Q) @
    -- I think the use of 'try' here *may* be very inefficient.
    pepaPrefix :: Parser ParsedComponent
    pepaPrefix = 
        do mTrans  <- pepaTrans
           symbol "."
           right   <- ((try pepaPrefix) <|> compFactor)
           return $ PrefixComponent mTrans right

    -- Pepa conditionals are simple if-thens, note no else.
    pepaConditional :: Parser ParsedComponent
    pepaConditional =
       do reserved "if"
          rCond         <- rateConditional
          reserved "then"
          right         <- ((try pepaPrefix) <|> compFactor)
          return $ CondBehaviour rCond right
       where
       rateConditional :: Parser BeCondition
       rateConditional =
          do ident    <- pepaRateId
             operator <- rateCondOperator
             rateExp  <- pepaRateExp 
             return $ BeCondition ident operator rateExp

       rateCondOperator :: Parser String
       rateCondOperator =
         ( symbol "<" <|> symbol ">" <|> symbol "=" <|> symbol ">="
          <|> symbol "<=" <|> symbol "<>")

    {- 
      A process tail is only afixable to a process name and allows
      the user to specify an array of processes. In addition to the number
      of processes to run in parallel the user can specify a list of actions
      on which the processes should cooperate. There are two styles for this
      @ P[3][a,b] @ is an array of 3 @P@ processes cooperating over
      the actions @a@ and @b@.
      @ P[3, {a, b}] @ is an alternative style for the same thing.
    -}
    processTail = do symbol "["
                     size <- natural
                     end  <- (do comma
                                 actions <- maybeOption $ braces pepaActionList
                                 symbol "]"
                                 return actions
                             ) <|>
                            (do symbol "]"
                                maybeOption $ squares pepaActionList
                            )
                     return (size, end)

    addProcessTail :: ParsedComponentId
                   -> Maybe ( Integer, Maybe [ParsedAction] )
                   -> ParsedComponent
    addProcessTail proc Nothing             = IdProcess proc
    addProcessTail proc (Just (size, acts)) = ProcessArray proc size acts
        
{-
  A parser for a pepa transitions.
  The first parser accepts the standard @(a, r)@ still of transition,
  however we can also put in just a single action name (which is
  lowercase so should not conflict with process identifiers) this is
  an immediate action and is shorthand for @(a, immediate)@
-}
pepaTrans :: Parser ParsedTrans
pepaTrans = parens ( do action <- pepaAction
                        priority <- maybeOption pepaPriority
		        comma
		        rate   <- pepaRate
	                return $ makeTrans action priority rate
	           )
            -- so a normal bracketed (a,r) style transition
            <|>
            -- or a single action identifier which will be performed
            -- as an immediate action, ag @ tau.P @
            ( do action <- pepaAction
                 priority <- maybeOption pepaPriority
                 return $ makeTrans action priority ( RateImmediate Nothing )
            )
            where
            makeTrans :: ParsedAction -> Maybe ParsedPriority
                            -> ParsedRate -> ParsedTrans
            makeTrans action priority rate =
               ParsedTrans { pepaTransAction   = action
                           , pepaTransPriority = maybe defaultPepaPriority id priority
                           , pepaTransRate     = rate
                           }
               

{- Parses a priority which may be affixed to the end of a (semi-markov) pepa action -}
pepaPriority :: Parser ParsedPriority
pepaPriority = squares intParser

{-
   Parses a rate. Unfortunately we have to use 'try'
   This is because both immediate and infty may actually be rate expressions.
   The simple case they might must be named longer, eg it might be the rate
   expression: @immediate_rate * r1@ if we didn't use 'try' then the
   @rateImm@ parser would consume the word @immediate@.
   Similarly for @rateTop@ it would consume @infty@  if confronted with
   the rate @infty_like_rate@.
-}
pepaRate :: Parser ParsedRate
pepaRate =
   (try rateImm)
   <|>
   (try rateTop)
   <|> ( pepaRateExp >>= (return . RateTimed) )
   where
   rateTop :: Parser ParsedRate
   rateTop = do (reserved "infty" <|> reserved "_")
                return RateTop

   rateImm :: Parser ParsedRate
   rateImm = do weight <- maybeOption weightParser
                reserved "immediate"
                return $ RateImmediate weight

   weightParser :: Parser Double
   weightParser = do i <- forgivingFloat
                     symbol ":"
                     return i


{-
  We build a parser for rate expressions using parsec's expression
  parser building library. Best really to look at the parsec
  documentation, but briefly the [-buildExpressionParser-] takes
  two arguments, the first is a table of operators where each row
  represents an operator precedence. The second argument is the parser
  for the basic terms, that is non-decomposable terms.

  This is quite good, it does not allow an immediate rate to be part of
  a binary rate expression, so for example one cannot have
  @ r * 2:immediate @ as this is obvious nonsense. However because the
  @condRate@ parser explictly calls 'pepaRateExp' rather than being built
  up from the @rateTable@ it is still possible to have something like:
  @ if P then 2:immediate else 1:immediate @
  which might be useful is perhaps a bit difficult to compile.
-}
pepaRateExp :: Parser ParsedRateExp
pepaRateExp = 
   (buildExpressionParser rateTable rateFactor )
   <?> "rate expression"
   where 

   rateTable   = [ [ op "*" (RateBexp "*") AssocLeft
                   , op "/" (RateBexp "/") AssocLeft 
                   ]
                 , [ op "+" (RateBexp "+") AssocLeft
                   , op "-" (RateBexp "-") AssocLeft 
                   ]
                 ]

   op :: String -> (ParsedRateExp -> ParsedRateExp -> ParsedRateExp)
        -> Assoc -> Operator Char () ParsedRateExp
   op opString f assoc = Infix (do reservedOp opString
                                   return f) assoc

   rateFactor  = parens pepaRateExp
                 <|> condRate
                 <|> rateIdent
                 <|> rateNum
                 <|> rateProcess
                 <?> "simple expression"
   rateIdent   = do ident <- pepaRateId
                    return (RateId ident Nothing)
   rateNum     = do num   <- forgivingFloat
                    return (RateNum num)
   rateProcess = do ident <- pepaComponentId
                    return $ RateProc ident
   condRate    = do reserved "if"
                    cond  <- condExp
                    reserved "then"
                    texp  <- pepaRateExp
                    reserved "else"
                    eexp  <- pepaRateExp
                    return $ RateCond cond texp eexp

   condExp     = do ident <- pepaComponentId
                    return $ PCondId ident


{-
  A pepa action should be either an action identifier or a tau of
  an action identifier, for now though only action identifiers
  are parsed.
-}
pepaAction :: Parser ParsedAction
pepaAction = do ident <- pepaActionId
		return (Action ident)

{-
  Pepa action list, separated by commas.
-}
pepaActionList :: Parser [ParsedAction]
pepaActionList = sepBy pepaAction comma


{-
  The parsers for the simple elements of a pepa file. By that I mean here is
  where we define what a rate identifier, and a process identifier is,
  for example.
 -}

{- 
   Rate identifiers begin with a lower case letter.
   We also take this opportunity to prefix the rate with a name.
   @todo{Rate identifiers should of course be stored as 
   'Language.Pepa.Syntax.QualifiedName' that way we can give good error
   messages (and come to think of it logging information).
   }
-}
pepaRateId :: Parser RateIdentifier
pepaRateId = 
   do namespace <- maybeOption $ do name <- upperId
                                    symbol "::"
                                    return name
      case namespace of
         Nothing -> do ident <- lowerId
                       return $ Unqualified ident
         Just p  -> do rest <- pepaRateId
                       return $ NameSpaceId p rest

{-
   But if it is the start of a definition then it can of course only
   be a non-namespaced name.
-}
pepaRateDefId :: Parser RateIdentifier
pepaRateDefId =
   do ident <- lowerId
      return $ Unqualified ident

{- Action identifiers begin with a lower case letter -}
pepaActionId :: Parser ActionIdentifier
pepaActionId = do ident <- lowerId
                  return $ Unqualified ident

{- Component identifiers begin with an upper case letter -}
pepaComponentId :: Parser ParsedComponentId
pepaComponentId = 
   do ident  <- upperId
      idTail <- maybeOption $ do symbol "::"
                                 rest <- pepaComponentId
                                 return rest
      return $ addTail ident idTail
   where
   addTail :: String -> Maybe ParsedComponentId -> ParsedComponentId
   addTail ident Nothing  = Unqualified ident
   addTail ident (Just p) = NameSpaceId ident p
		  

{-
   Similarly as for rate definitions, for component definitions the
   name must be a non-namespaced one.
-}
pepaProcessDefId :: Parser ParsedComponentId
pepaProcessDefId =
   do ident <- upperId
      return $ Unqualified ident
                    

{-
  This subsection contains a few generally useful parsers.
  These include things like a lowercase identifier. 
  The main parser can then use these by for example setting a rate identifier
  to be a lowercase identifier.
 -}
alphaNumUnderScore :: Parser Char
alphaNumUnderScore = alphaNum <|> char '_'

lowerId :: Parser String
lowerId = do firstChar <- lower
             restChars <- many alphaNumUnderScore
             whiteSpace
             return (firstChar : restChars)

upperId :: Parser String
upperId = do firstChar <- upper
             restChars <- many alphaNumUnderScore
             whiteSpace
             return (firstChar : restChars)

allUpperId :: Parser String
allUpperId = do ident <- many1 upper
                whiteSpace
                return ident

floatNumber :: Parser Double
floatNumber = StdToken.float lexer

{-
  [-forgivingFloat-] is the same as [-floatNumber-] except that as
  a short hand for a whole number one can leave off the decimal
-}
forgivingFloat :: Parser Double
forgivingFloat = do eitherNum <- naturalOrFloat
		    return $ doubleOfEitherNum eitherNum
	         where doubleOfEitherNum :: Either Integer Double -> Double
		       doubleOfEitherNum (Left i)  = fromInteger i
		       doubleOfEitherNum (Right d) = d

{-
  A comma separated list of floating point numbers
-}
floatNumberList :: Parser [Double]
floatNumberList = sepBy1 floatNumber comma


{-
  A comma separated list of floating point numbers, may not
  seem necessary but for example there are cases where there
  are several options all of which begin with a floating point
  number, if one of the options is a list then we have to parse
  the rest of the list with this.
  For example, the [- --rates -] option, accepts either
  [- r=1.2,1.3,1.4,1.5 -]
  or
  [- r=1.2..1.5:0.1 -]
  In both cases it starts with an initial floating point number,
  then we have to decide whether we parse a list if a comma
  follows or the range and step if two dots follow.
-}
commaFloatNumberList :: Double -> Parser [Double]
commaFloatNumberList first = do comma
				numbers <- floatNumberList
				return (first : numbers)

{-
  This final parser combinator should be in the parsec library
  but I can't find exactly this.
-}
maybeOption :: GenParser tok st a -> GenParser tok st (Maybe a)
maybeOption p = ( do value <- try p
		     return (Just value)
		) <|> do { return Nothing }


{-
  Token Parser Stuff !!!
  The token parser essentially allows us to forget about the
  whitespace.

  The following definition of 'lexer' is required for the token 
  parser stuff to work.
  What I should do is comment this better.

  Note: Although command line options are less likely to contain spaces
  since a space generally separates out a command line option it
  is still possible that they contain spaces by quoting an option
  for example we could have
  @ ipc --rates "r = 1.2 .. 2.1 : 0.1" ... @
  as a command line, hence there are spaces within the arguments
  to the @--rates @ option.
-}
lexer :: StdToken.TokenParser ()
lexer = 
    PrvToken.makeTokenParser langDef
    where 
    langDef1  = PrvToken.translateLanguageDef haskellStyle
    langDef   = langDef1 { PrvToken.reservedOpNames = reservedOperators
			 , PrvToken.reservedNames   = pepaNames
			 , PrvToken.commentLine     = [ "%", "//" ]
			 , PrvToken.commentStart    = "{-"
			 , PrvToken.commentEnd      = "-}"
			 }
    pepaNames = [ "else"
		, "if"
		, "infty"
		, "then"
		, "immediate"
		, "_"
		--, "stop" 
		]
    reservedOperators = [ "*"
			, "/"
			, "+"
			, "-"
			  
			  -- Probe operators
			, "?"
			]


{-
  For convenience we alias some of the common token parsers.
  Along with a couple commented out because we do not use them
  currently but may at some point.
-}
whiteSpace :: CharParser () ()
whiteSpace = StdToken.whiteSpace lexer


--lexeme     :: CharParser () a -> CharParser () a
--lexeme     = StdToken.lexeme lexer

symbol     :: String -> CharParser () String
symbol     = StdToken.symbol lexer

natural :: CharParser () Integer
natural    = StdToken.natural lexer

naturalOrFloat :: CharParser () (Either Integer Double)
naturalOrFloat = StdToken.naturalOrFloat lexer

integer :: CharParser () Integer
integer    = StdToken.integer lexer

intParser :: CharParser () Int
intParser  = do i <- integer
                return $ fromInteger i


parens     :: CharParser () a -> CharParser () a
parens     = StdToken.parens lexer

angles     :: CharParser () a -> CharParser () a
angles     = StdToken.angles lexer

squares    :: CharParser () a -> CharParser () a
squares    = StdToken.squares lexer

braces     :: CharParser () a -> CharParser () a
braces     = StdToken.braces lexer

--semi       :: CharParser () String
--semi       = StdToken.semi lexer

comma       :: CharParser () String
comma        = StdToken.comma lexer

--colon       :: CharParser () String
--colon       = StdToken.colon lexer

--identifier :: CharParser () String
--identifier = StdToken.identifier lexer

reserved   :: String -> CharParser () ()
reserved   = StdToken.reserved lexer

reservedOp :: String -> CharParser () ()
reservedOp = StdToken.reservedOp lexer

