{-|
   A module of  combinators for building
   non-deterministic finite state machines.
   Note these combinators are for nfas in relation to stochastic
   probes. These are therefore ultimately cyclic machines though
   at the stage these combinators are used that is not the case.

   If you are looking for a generic finite automata library, this
   is probably not it, though may be easier to amend than others. 
-}
module Language.Pepa.Probes.StateMachines.NfaCombinators
   ( IntNfa

   , nfaOr
   , nfaThen
   , nfaStar
   , nfaQuestion
   , nfaReset
   )
where

{- Standard library module imports -}
import Prelude hiding
   ( map
   , any
   )
import Data.Foldable
   ( any )
import Data.Maybe
   ( mapMaybe )
import Data.Set
   ( Set
   , member
   , insert
   , union
   , singleton
   , size
   , toList
   , fromList
   , map
   )
{- Local module imports -}
import Language.Pepa.Probes.StateMachines.NfaTypes
   ( Nfa       ( .. )
   , NfaState  ( .. )
   , NfaLabel  ( .. )

   , Move      ( .. )
   )
{- End of module imports -}

{-|
   The type of a non-deterministic finite state automata with integers
   as their states.
-}
type IntNfa a = Nfa Int a
-------------------------------------------------------------------------- 
--									--
--	Combinators for machines use to build them			--
--									--
-------------------------------------------------------------------------- 

{- Of course this *should* have type:
@
nfaOr :: (NfaState state, NfaLabel label) => 
         Nfa state label        -- The left nfa
      -> Nfa state label        -- The right nfa
      -> Nfa state label
@
-}
{-|
   Creates an nfa which can perform as the first input nfa
   or the second nfa.
-}
nfaOr :: (NfaLabel a) => IntNfa a -> IntNfa a -> IntNfa a
nfaOr (NFA states1 moves1 _start1 _finish1)
      (NFA states2 moves2 _start2 _finish2) =
   NFA allStates allMoves newStart $ singleton ( m1 + m2 + 1 )
   where
   allStates = union states1' $ union states2' newStates
   allMoves  = union moves1' $ union moves2' newMoves

   newStart  = minimal

   m1 = size states1
   m2 = size states2
   states1' = map (renumber 1) states1
   states2' = map (renumber (m1+1)) states2
   newStates = fromList [0,(m1+m2+1)]
   moves1'  = map (renumber_move 1) moves1
   moves2'  = map (renumber_move (m1+1)) moves2
   newMoves = fromList [ Emove 0 1 
                       , Emove 0 (m1+1)
                       , Emove m1 (m1+m2+1)
                       , Emove (m1+m2) (m1+m2+1)
                       ]

{- I think there is a bug in this, shouldn't it be able to go from
   *any* of the accepting states of finish1 to start2 ?

   Okay when I've got time I can try the following:
   @ k = size states1 @
   @newMoves = map (\x -> Emove x (start2 $ renumber k)) finish1@
   Oh actually we definitely do *not* want to do it like this, at least
   not for our finite state machines, because if we are using 'accepting'
   to mean immediate communication states then we definitely don't want
   to arbitrarily jump out of those states.
-}
{-|
   Performs as the first input nfa and *then* as the second.
-}
nfaThen :: (NfaLabel a) => IntNfa a -> IntNfa a -> IntNfa a
nfaThen (NFA states1 moves1 start1 _finish1) 
        (NFA states2 moves2 _start2 finish2) =
   NFA
   (union states1 states2')
   (union moves1 moves2')
   start1
   finish2'
   where
   states2' = map (renumber k) states2
   moves2'  = map (renumber_move k) moves2
   finish2' = map (renumber k) finish2
   k = (size states1) - 1


{-|
   Performs as the input nfa any number of times including zero times.
   Most useful if the input nfa is not a single activity since in that
   case it is the same as no nfa at all since the self-loops will add
   the ability to do any number of that activity in any state.
-}
nfaStar :: (NfaLabel a) => IntNfa a -> IntNfa a
nfaStar (NFA states moves _start _finish) =
   NFA
   (states' `union` newstates)
   (moves' `union` newmoves)
   0
   (singleton (m+1))
   where
   m         = size states
   states'   = map (renumber 1) states
   newstates = fromList [ 0 , m+1 ]
   moves'    = map (renumber_move 1) moves
   newmoves  = fromList [ Emove 0 1
                        , Emove m 1
                        , Emove 0 (m+1)
                        , Emove m (m+1)
                        ]

{-| 
  A bit similar to 'nfaStar' except that it will do the machine
  once or zero times.
-}
nfaQuestion :: (NfaLabel a) => IntNfa a -> IntNfa a
nfaQuestion (NFA states1 moves1 start1 _finish1) =
   NFA (insert newEnd states1)
       (union moves1 newMoves)
       start1
       (singleton newEnd)
   where
   k        = size states1 - 1
   newEnd   = k + 1
   newMoves = fromList [ Emove start1 newEnd, Emove k newEnd ]

{-|
   Takes an nfa and a label and resets the nfa to the start state
   at every state in which the given label cannot be performed.
-}
nfaReset :: (NfaLabel l) => l -> IntNfa l -> IntNfa l
nfaReset action (NFA states moves start finish) =
   (NFA states newMoveSet start finish )
   -- For every move if we cannot currently perform
   -- 'action' then move from 'action' to the start state.
   where
   newMoveSet = union moves newMoves
   newMoves   = mapMaybeSet resetAction states

   -- resetAction :: Int -> Maybe (Move Int l)
   resetAction i
      -- if we have finished we cannot reset
      | member i finish           = Nothing
      | any iPerformsAction moves = Nothing
      | otherwise                 = Just $ Move i action 0 -- start
      where
      -- iPerformsAction :: Move Int l -> Bool
      iPerformsAction (Move j lab _) = (j == i) && (lab == action)
      iPerformsAction _              = False

-------------------------------------------------------------------------- 
--      Auxilliary functions used in the building cominbators above     --
--      regular expressions.                                            --
-------------------------------------------------------------------------- 
renumber_move :: (NfaState a, NfaLabel b) => Int -> Move a b -> Move a b
renumber_move k (Move s1 c s2) =
   Move (renumber k s1) c (renumber k s2)
renumber_move k (Emove s1 s2)  =
   Emove (renumber k s1) (renumber k s2)

-- A temporary mapMaybe for sets
mapMaybeSet :: (Ord a, Ord b) => (a -> Maybe b) -> Set a -> Set b
mapMaybeSet f = fromList . (mapMaybe f) . toList