{-| 
   This module translates an Srmc model into a Pepa model.
-}
module Language.Pepa.Srmc.SrmcPepa
   ( TranslatedModel     ( .. )
   , RateArray -- Maybe this should be defined elsewhere
   , translateSrmc
   )
where

{- Standard Libraries modules imported -}
import Data.List
   ( partition )
{- External Library modules imported -}
{- Local modules imported imported -}
import Language.Pepa.Syntax
   ( QualifiedName     ( .. )
   , RateIdentifier 
   , ParsedComponentId 
   , ShowOrig          ( .. )
   , ParsedModel       ( .. )
   , ParsedComponent   ( .. )
   , BeCondition       ( .. )
   , ParsedTrans       ( .. )
   , ParsedRate        ( .. )
   , ParsedRateExp     ( .. )
   , ParsedCond        ( .. )
   )   
import Language.Pepa.Srmc.Syntax
   ( SrmcModel         ( .. )
   , SrmcDef           ( .. )
   )
import Language.Pepa.Utils
   ( mapSnd
   , permuteDefinitionArrays
   )
import Ipc.Cli
   ( CliOpt
   , getProcessNumber
   )
import Smc.Cli
   ( SmcOptions )

import Language.Pepa.Analysis.Analysis
   ( analyseModel )
import Language.Pepa.Transform.Simplify
   ( removeUnusedDefinitions
   , relegateConditionalBehaviour
   , removeAliases
   )
{- End of imports -}



{- 
   This function provides the simplifying of the PEPA models
   which we must do. Most of this isn't really necessary
   but is done for neatness. The old ipc as well cannot
   cope with alias statements so we need to remove those
   (although once the new ipc based on ipclib is in use
   this will be less necessary and could perhaps be looked
   at).

   Note that the report must be on the simplified model
   since it may contain extra definitions after the simiplification
   such as lifted prefixes.

   Argh, I'd like to call 'Languages.Pepa.Transform.Simplify.simplfy'
   instead of 'Languages.Pepa.Transform.Simplify.removeAliases' but
   unfortunately this lifts the rate expressions which has the unfortunate
   side-effect that the analysis which checks which rates are used will
   see rate identifiers used in the lifted rate expressions  as being
   used, even though their rate definitions are not.
   Basically we definitely need to fix the analysis of rates used
   but until then we'll do this instead.
-}
simplifyModel :: ParsedModel -> ParsedModel
simplifyModel model =
   removeUnusedDefinitions report simplifiedModel
   where
   report          = analyseModel simplifiedModel
   simplifiedModel = removeAliases $ relegateConditionalBehaviour model


{-
   The type of a rate array is the type of a rate identifier mapped
   to a list of possible rate expressions which will/could be
   ranged over in a parameter sweep computation.
-}
type RateArray = (RateIdentifier, [ ParsedRateExp ] )

{-|
   The type returned by the translation function.
   Basically we wish to return the translated model
   as well as all the rates which way may range over
   in parameter sweep computations.
-}
data TranslatedModel =
   TModel ParsedModel [ RateArray ]

{-|
   The main exported function, this takes in an Srmc model and
   translates it into a list of pepa models.
   The pepa models correspond to the different ways in which the
   processes can be combined together. So for example if we had
   a simple srmc model in which the only two set of processes
   definitions were:
   @P = { P1, P2 } @
   @Q = { Q1, Q2 } @
   Then there would be four output pepa models each setting
   the first being
   @ P = P1 ; Q = Q1@
   the second being
   @ P = P2 ; Q = Q1@
   and so on.
-}
translateSrmc :: [ CliOpt SmcOptions ] -> SrmcModel -> [ TranslatedModel ]
translateSrmc options (SrmcModel sDefs mainComp) =
   map (makeTModel . makeModel) processNumbers
   where
   -- Makes the pepa model by selecting the appropriate process definitions
   -- based on the given number. This model is then simplified to remove
   -- any unused rate and process definitions, since by selecting some
   -- processes this will mean many definitions are redundant and we do not
   -- wish to, for example, range over a rate which is not used in that
   -- model.
   makeModel :: Int -> ParsedModel
   makeModel number =
      simplifyModel model
      where
      model           = ParsedModel rateDefs processDefs scopedMainComp
      processDefs     = uniqueProcessDefs ++
                        map (selectSetDef number) processArrays

   -- To make the translated model we must add in the rate arrays
   -- however we only wish to include those rate arrays which are
   -- actually used in the model. Since the simplification of the
   -- model included removing unused definitions we can make the
   -- assumption that if a rate is defined then it is used
   -- (even if not then it is a valid rate to be ranging over).
   -- So essentially then this function filters all the rate arrays
   -- to include only those that are used in the given model and
   -- combines those filtered rate arrays with the given model to
   -- return the translated model.
   -- Note that the actual model is not modified.
   makeTModel :: ParsedModel -> TranslatedModel
   makeTModel model@(ParsedModel rDefs _ _) =
      TModel model usedRateArrays
      where
      -- Note we use the original rate arrays and not their permuted forms
      -- since this is how it we wish it to appear in the condor file.
      -- The later run of ipc will permute it itself.
      usedRateArrays  = filter isUsed origRateArrays
      usedRateIdents  = map fst rDefs
      isUsed :: RateArray -> Bool
      isUsed = ((flip elem usedRateIdents) . fst)

   processNumbers     = [0 .. processArrayLength]
   processArrayLength = case processArrays of
                           []      -> 0
                           (h : _) -> (length $ snd h) - 1
   

   scopedMainComp     = scopeProcess scope mainComp
   uniqueProcessDefs  = (map selectSetDefHead processSpecs)
   rateDefs           = (map selectSetDefHead rateSpecs)
                         ++ (map selectRateDef rateArrays)

   -- split the rates into those with a single definition and those with
   -- a set of possible values. 
   (rateSpecs, origRateArrays) = partition ((==1) . (length . snd)) $ fst cDefs
   rateArrays                  = permuteDefinitionArrays origRateArrays

   -- Do the same for the process definitions, split them into single definitions
   -- and true sets.
   (processSpecs, processArrays1) = partition ((==1) . (length . snd)) $ snd cDefs
   processArrays                  = permuteDefinitionArrays processArrays1

   -- Create the scope in which to scope the main composition and the
   -- top level of definitions.
   mainNameSpace = Unqualified "Main"
   scope         = createDefinitionsScope mainNameSpace emptyScope sDefs

   -- Use the top level scope to convert the top-level definitions.
   cDefs         = convertDefinitions scope sDefs

   -- The process number is used to select which specific 
   -- rate definition from the definitions sets. For srmc generally we will
   -- later be ranging over the rate definition sets so it matters little
   processNumber =
      case getProcessNumber options of
         Nothing -> 0
         Just i  -> i

   -- select just one rate or process from a set definition since set
   -- definitions are not allowed in ordinary pepa
   selectSetDefHead :: (a , [ b ]) -> (a, b)
   selectSetDefHead = mapSnd head

   selectSetDef :: Int -> (a, [ b ]) -> (a, b)
   selectSetDef number = mapSnd $ (flip (!!)) number

   selectRateDef :: (a, [ b ]) -> (a, b)
   selectRateDef = selectSetDef processNumber


{- The type of a namespace is quite straight-forward -}
type NameSpace = QualifiedName

{-
   A scoping maps rate and process identifiers to their drawn out
   names. So that for example we may have the mapping
   @ r --> PONE__r @
   This mapping allows us to change all occurrences of @r$ to @PONE__R@
-}
type ScopeEntry = ( String, QualifiedName ) 
type Scope      = [ ScopeEntry ]

emptyScope :: Scope
emptyScope = []

{-
prefixQualified :: QualifiedName -> QualifiedName -> QualifiedName
prefixQualified prefix n = Unqualified $ concat [ showOrig prefix
                                                , "__"
                                                , showOrig n
                                                ]
-}

attachModuleName :: QualifiedName -> QualifiedName -> QualifiedName
attachModuleName (NameSpaceId longid qname) ident =
   NameSpaceId longid $ attachModuleName qname ident
attachModuleName qname ident =
   NameSpaceId (showOrig qname) ident


{-
   Takes a set of definitions and creates a scope in which they should
   be themselves scoped.
   It may seem strange that we separate out the functionality of
   'createDefinitionsScope' and 'convertDefinitions' however since the
   main system component must be scoped we require a way to calculate
   the scope without converting any definitions.
-}
createDefinitionsScope :: NameSpace -> Scope -> [ SrmcDef ] -> Scope
createDefinitionsScope  prefix scope sDefs =
   newScope ++ scope
   where
   newScope = map makeSrmcMapping sDefs

   makeSrmcMapping :: SrmcDef -> ScopeEntry
   makeSrmcMapping (ServiceDef ident _)  =
      ( showOrig ident, attachModuleName prefix ident )
   makeSrmcMapping (RateSet ident _)     =
      ( showOrig ident, attachModuleName prefix ident )
   makeSrmcMapping (ProcessSet ident _)  =
      ( showOrig ident, attachModuleName prefix ident )


type RateSetDef    = RateArray -- ( RateIdentifier, [ ParsedRateExp ] )
type ProcessSetDef = ( ParsedComponentId, [ ParsedComponent ] )
type ConvertedDefs = ( [ RateSetDef ], [ ProcessSetDef ] )

{- 
   Converts a set of rate, process and service definitions according
   to the given scope and the enclosing namespace.

   Note in all the definitions we scope the identifier, this will
   always succeed since we added all those into the scope when
   we called 'createDefinitionsScope'.
-}
convertDefinitions :: Scope -> [ SrmcDef ] -> ConvertedDefs
convertDefinitions thisScope sDefs =
   ( rateDefs, processDefs )
   where
   rateDefs          = concat $ map fst definitions
   processDefs       = concat $ map snd definitions
   definitions       = map convertSDef sDefs

   {-
      So now we must recursively convert the nested service
      definitions.
   -}
   convertSDef :: SrmcDef -> ConvertedDefs
   convertSDef (ServiceDef ident ssDefs) =
      convertDefinitions newScope ssDefs
      where
      newPrefix = scopeIdentifier thisScope ident
      newScope  = createDefinitionsScope newPrefix thisScope ssDefs
   convertSDef (RateSet ident rates)                   =
      ( [ (newIdent, newRates) ], [] )
      where
      newIdent = scopeIdentifier thisScope ident
      newRates = map (scopeRateExpression thisScope) rates
   convertSDef (ProcessSet ident procs)                =
      ( [], [ ( newIdent, newProcesses ) ] )
      where
      newIdent     = scopeIdentifier thisScope ident
      newProcesses = map (scopeProcess thisScope) procs

scopeIdentifier :: Scope -> QualifiedName -> QualifiedName
scopeIdentifier scope (NameSpaceId namespace qname) =
   case lookup namespace scope of
      Nothing     -> error "Namespace used out of scope"
      Just longid -> attachModuleName longid qname
scopeIdentifier scope ident                         =
   case lookup (showOrig ident) scope of
      Nothing     -> error "Used name not in scope"
      Just longid -> longid


scopeRate :: Scope -> ParsedRate -> ParsedRate
scopeRate _scope r@(RateTop)         = r
scopeRate _scope r@(RateImmediate _) = r
scopeRate scope  (RateTimed expr)    = 
   RateTimed $ scopeRateExpression scope expr

{-
   Scopes a rate expression. During the conversion rate names
   and process names will be renamed with a prefix indicating
   in which name space they are in. Hence rate expressions which
   use those scoped rate and process names must change the names
   to which they refer.
-}
scopeRateExpression :: Scope -> ParsedRateExp -> ParsedRateExp
scopeRateExpression _scope r@(RateNum _)             = r
scopeRateExpression scope  (RateId ident mR)         =
   RateId (scopeIdentifier scope ident) mR
scopeRateExpression scope (RateBexp oper left right) =
   RateBexp oper (scopeRateExpression scope left)
                 (scopeRateExpression scope right)
scopeRateExpression scope (RateProc ident)           =
   RateProc $ scopeIdentifier scope ident
scopeRateExpression scope (RateCond cond left right) =
   RateCond (scopeRateCondition cond)
            (scopeRateExpression scope left)
            (scopeRateExpression scope right)
   where
   scopeRateCondition :: ParsedCond -> ParsedCond
   scopeRateCondition (PCondId ident)                =
      PCondId $ scopeIdentifier scope ident
   scopeRateCondition (PCondSum leftcond rightcond)  =
      PCondSum (scopeRateCondition leftcond)
               (scopeRateCondition rightcond)
   scopeRateCondition (PCondRate rIdent oper rExpr)  =
      PCondRate (scopeIdentifier scope rIdent)
                oper
                (scopeRateExpression scope rExpr)
scopeRateExpression _scope r@(RateDisabled)         = r


{-
   Scopes a process including the rate expressions contained within
-}
scopeProcess :: Scope -> ParsedComponent -> ParsedComponent
scopeProcess scope  (IdProcess ident)                    =
   IdProcess $ scopeIdentifier scope ident
scopeProcess _scope p@(StopProcess)                      = 
   p
scopeProcess scope (PrefixComponent trans next) =
   PrefixComponent newTrans (scopeProcess scope next)
   where
   newTrans = trans { pepaTransRate = newRate }
   newRate  = scopeRate scope $ pepaTransRate trans
scopeProcess scope (CondBehaviour cond next)             =
   CondBehaviour (scopeConditionalBehaviour scope cond)
                 (scopeProcess scope next)
scopeProcess scope (ComponentSum left right)             =
   ComponentSum (scopeProcess scope left)
                (scopeProcess scope right)
scopeProcess scope (Cooperation left actions right)      =
   Cooperation (scopeProcess scope left)
               actions
               (scopeProcess scope right)
scopeProcess scope (ProcessArray ident actions size)     =
   ProcessArray (scopeIdentifier scope ident)
                actions
                size
scopeProcess scope (Hiding comp actions)                 =
   Hiding (scopeProcess scope comp) actions

scopeConditionalBehaviour :: Scope -> BeCondition -> BeCondition
scopeConditionalBehaviour scope (BeCondition rateId op rateExp) =
   BeCondition (scopeIdentifier scope rateId)
               op
               (scopeRateExpression scope rateExp)
