{-| This module implements simplifying transformations over
    parsed pepa models
-}
module Language.Pepa.Transform.Simplify
   ( simplify
   , liftComplex
   , liftRateExprs
   , removeAliases
   , expandProcessArrays
   , removeParallelDefs
   , removeConditionalBehaviour
   , removeUnusedDefinitions
   , relegateConditionalBehaviour
   , removeHidingUnqualifiedModel
   )
where

{- Standard Libraries modules imported -}
import Data.List
   ( mapAccumL
   , unionBy
   )
import Data.Maybe
   ( mapMaybe
   , fromMaybe
   )
{- External Library modules imported -}
{- Local modules imported imported -}
import Ipc.Cli
   ( CliOpt    ( .. )
   , containsHideNonCoop
   )
import Language.Pepa.Syntax
   ( ParsedModel       ( .. )
   , ProcessDef
   , RateSpec
   , ParsedComponentId 
   , ParsedComponent   ( .. )
   , BeCondition       ( .. )
   , ParsedRate        ( .. )
   , ParsedRateExp     ( .. )
   , ParsedCond        ( .. )
   , ParsedValue 
   , ParsedTrans       ( .. )
   , ParsedAction      ( .. )
   , ActionIdentifier
   , QualifiedName     ( .. )
   , qualifyQName
   , ShowOrig          ( .. )
   )
import Language.Pepa.Transform.NameSupply
   ( NameSupply
   , initialNameSupply
   , getName
   )
import Language.Pepa.Utils
    ( mapSnd )

import Language.Pepa.Analysis.Analysis 
    ( AnalysisReport      ( .. )
    , analyseModel
    )
import Language.Pepa.Transform.HideAll
   ( hideAllActions )
{- End of imports -}


{-|
   Applies all the simplifications to a model.
   In the future this should actually take into account the
   command-line options.
   Currently they are all ignored though.
-}
simplify :: [ CliOpt a ] -> ParsedModel -> ParsedModel
simplify options = 
   liftComplex
   . liftRateExprs
   . removeHiding
   . hideNonCooperating
   . qualifyModel
   . removeAliases
   . removeParallelDefs
   where
   -- Obviously the test should be about the existance of a command-line
   -- option.
   hideNonCooperating :: ParsedModel -> ParsedModel
   hideNonCooperating model
      | containsHideNonCoop options = 
         hideAllActions (analyseModel model) model
      | otherwise                   = model


{-|
    lift complex -- This tranformation will lift complex sub-process forms
    to become definitions of their own.
    For example:
    @ P = (a, r).(b, r).Q ; @
    becomes
    @ P  = (a, r).P1 ;
      P1 = (b, r).Q  ;
    @
-}
liftComplex :: ParsedModel -> ParsedModel
liftComplex (ParsedModel rDefs pDefs system) =
    (ParsedModel rDefs newDefinitions system)
    where
    newDefinitions = aliasDefList initialNameSupply pDefs

    -- Possibly a little tricky to understand but it's pretty simple really
    -- 'mapAccumL' applies 'aliasDef' to each of process definitions, with
    -- the name supply as the accumulator. So basically each process definition
    -- is processed with the name supply accumulated so far. The processing
    -- of each definition will, along with the accumulated name supply return
    -- a list of new definitions, so we end up with a list of list of process
    -- definitions, so we concatenate them to give us our final list of
    -- process definitions.
    aliasDefList :: NameSupply -> [ ProcessDef ] -> [ ProcessDef ]
    aliasDefList ns = concat . snd . (mapAccumL aliasDef ns)

    aliasDef :: NameSupply -> ProcessDef -> ( NameSupply, [ ProcessDef ] )
    aliasDef renamer (ident, process) =
        (newRenamer, (ident, newProcess) : newDefs)
        where
        (newRenamer, newProcess, newDefs) = aliasProcess renamer process

    aliasProcess :: NameSupply -> ParsedComponent
                -> ( NameSupply, ParsedComponent, [ ProcessDef ] )
    aliasProcess renamer proc@(StopProcess)                         =
        (renamer, proc, [])
    aliasProcess renamer proc@(IdProcess _)                         =
        (renamer, proc, [])
    aliasProcess renamer proc@(PrefixComponent _ (IdProcess _))     =
        (renamer, proc, [])
    aliasProcess renamer (PrefixComponent trans nextProcess)        =
        (newNs, newProcess, newDef : processDefs)
        where
        newDef        = (newIdent, newNext)
        newIdent      = Unqualified $ snd newName
        newProcess    = PrefixComponent trans $ IdProcess newIdent
        (newNs, 
         newNext,
         processDefs) = aliasProcess (fst newName) nextProcess
        newName       = getName renamer "Prefix_"
    aliasProcess renamer proc@(CondBehaviour _ (IdProcess _))       =
        (renamer, proc, [])
    aliasProcess renamer (CondBehaviour cond nextProcess)           =
        (newNs, newProcess, newDef : processDefs)
        where
        newDef        = (newIdent, newNext)
        newIdent      = Unqualified $ snd newName
        newProcess    = CondBehaviour cond $ IdProcess newIdent
        (newNs, 
         newNext,
         processDefs) = aliasProcess (fst newName) nextProcess
        newName       = getName renamer "CondB_"

    {-
      Component sums are a bit awkward since we do not wish to introduce
      needless aliases such @P = S@, therefore we have to check if either
      of the branches are already an identifier process. There are therefore
      four cases, both are identifiers, the left is, the right is and neither
      are identifiers.
    -}
    aliasProcess renamer proc@(ComponentSum (IdProcess _)  
                                            (IdProcess _))          =
       ( renamer, proc, [] )
    aliasProcess renamer (ComponentSum left right@(IdProcess _))    =
       ( leftNs
       , ComponentSum (IdProcess newIdent) right
       , newDef : leftDefs
       )
       where
       newDef      = (newIdent, leftProc)
       newIdent    = Unqualified $ snd newName
       newName     = getName renamer "LeftCompSum_"
       (leftNs,
        leftProc,
        leftDefs)  = aliasProcess (fst newName) left
    aliasProcess renamer (ComponentSum left@(IdProcess _) right)    =
       ( rightNs
       , ComponentSum left $ IdProcess newIdent
       , newDef : rightDefs
       )
       where
       newDef      = (newIdent, rightProc)
       newIdent    = Unqualified $ snd newName
       newName     = getName renamer "RightCompSum_"
       (rightNs,
        rightProc,
        rightDefs) = aliasProcess (fst newName) right
    aliasProcess renamer (ComponentSum left right)                  =
        ( rightNs
        , ComponentSum (IdProcess leftIdent) (IdProcess rightIdent)
        , leftDef : rightDef : (leftDefs ++ rightDefs)
        )
        where
        leftDef     = (leftIdent, leftProc)
        rightDef    = (rightIdent, rightProc)
        
        leftIdent   = Unqualified $ snd leftName
        rightIdent  = Unqualified $ snd rightName

        leftName    = getName renamer "LeftCompSum_"
        rightName   = getName (fst leftName) "RightCompSum"

        (leftNs,
         leftProc,
         leftDefs)  = aliasProcess (fst rightName) left
        (rightNs,
         rightProc,
         rightDefs) = aliasProcess leftNs right
    -- Now all of the parallel operators should not contain any prefix operator.
    -- so they can be returned as is. We may revise this if we decide that we
    -- want to allow @(a,r).P <> (b,r).P@ which in some sense is not ludicrous.
    aliasProcess renamer proc@(Cooperation _ _ _)                   =
        (renamer, proc, [])
    aliasProcess renamer proc@(ProcessArray _ _ _)                  =
        (renamer, proc, [])
    aliasProcess renamer proc@(Hiding _ _)                          =
        (renamer, proc, [])

{- lift rate expressions -}
liftRateExprs :: ParsedModel -> ParsedModel
liftRateExprs (ParsedModel rDefs pDefs system) =
   ParsedModel (rDefs ++ newRateDefs) newProcessDefs system
   where
   newRateDefs    = concat $ fst newDefinitions
   newProcessDefs = snd newDefinitions
   newDefinitions = unzip $ liftDefList initialNameSupply pDefs

   -- Possibly a little tricky to understand but it's pretty simple really
   -- 'mapAccumL' applies 'liftDef' to each of process definitions, with
   -- the name supply as the accumulator. So basically each process definition
   -- is processed with the name supply accumulated so far. The processing
   -- of each definition will, along with the accumulated name supply return
   -- a list of new definitions, so we end up with a list of list of process
   -- definitions, so we concatenate them to give us our final list of
   -- process definitions.
   liftDefList :: NameSupply -> [ ProcessDef ]
               -> [ ( [ RateSpec ], ProcessDef ) ]
   liftDefList ns = snd . (mapAccumL liftDef ns)

   liftDef :: NameSupply -> ProcessDef 
           -> ( NameSupply, ( [ RateSpec ], ProcessDef ) )
   liftDef renamer (ident, process) =
       (newRenamer, ( newRDefs, (ident, newProcess)) )
       where
       (newRenamer, 
        newProcess, 
        newRDefs)  = liftProcess renamer process
   
   liftProcess :: NameSupply -> ParsedComponent
               -> (NameSupply, ParsedComponent, [ RateSpec ])
   liftProcess renamer proc@(StopProcess)                                    =
       (renamer, proc, [])
   liftProcess renamer proc@(IdProcess _)                                    =
       (renamer, proc, [])
   liftProcess renamer (PrefixComponent trans nextP)                         =
      case pepaTransRate trans of
         RateTop                -> nonLifted
         RateImmediate _        -> nonLifted
         RateTimed (RateId _ _) -> nonLifted
         RateTimed rExp         -> 
            (newLiftedR, liftedP, rDef : newLiftedRdefs)            
            where
            liftedP      = PrefixComponent newTrans newLiftedNext
            (newLiftedR,
             newLiftedNext,
             newLiftedRdefs) = liftProcess (fst newName) nextP
            
            newIdent     = Unqualified $ snd newName
            newName      = getName renamer "liftedRate_"
            newTrans     = trans { pepaTransRate = newRate }
            newRate      = RateTimed $ RateId newIdent $ Just rExp
            rDef            = (newIdent, rExp)
      where
      nonLifted    = (newRenamer, nonLiftedP, newRDefs)
      nonLiftedP   = PrefixComponent trans newNextP
      (newRenamer,
       newNextP,
       newRDefs)   = liftProcess renamer nextP

{-
   liftProcess renamer (PrefixComponent (t, RateTop) nextP)                  =
       (newRenamer, newProc, newRDefs)
       where
       newProc      = PrefixComponent (t, RateTop) newNextP
       (newRenamer,
        newNextP,
        newRDefs)   = liftProcess renamer nextP
   liftProcess renamer (PrefixComponent (t, RateImmediate weight) nextP)     =
       (newRenamer, newProc, newRDefs)
       where
       newProc      = PrefixComponent (t, RateImmediate weight) newNextP
       (newRenamer,
        newNextP,
        newRDefs)   = liftProcess renamer nextP
   liftProcess renamer (PrefixComponent (t, RateTimed (RateId n mr)) nextP)  =
       (newRenamer, newProc, newRDefs)
       where
       newProc      = PrefixComponent (t, RateTimed $ RateId n mr) newNextP
       (newRenamer,
        newNextP,
        newRDefs)   = liftProcess renamer nextP
   liftProcess renamer (PrefixComponent (t, RateTimed rExp) nextP)           =
       (newRenamer, newProcess, newRDef : newRDefs)
       where
       newRDef      = (newIdent, rExp)
       newProcess   = PrefixComponent newTrans newNextP
       newTrans     = (t, RateTimed $ RateId newIdent $ Just rExp)
       newIdent     = Unqualified $ snd newName
       newName      = getName renamer "liftedRate_"

       (newRenamer,
        newNextP,
        newRDefs)   = liftProcess (fst newName) nextP
   -}
   liftProcess renamer (CondBehaviour cond nextP)                            =
      (newRenamer, newProc, newRDefs)
      where
      newProc       = CondBehaviour cond newNextP
      (newRenamer,
        newNextP,
        newRDefs)   = liftProcess renamer nextP
   liftProcess renamer (ComponentSum left right)                             =
       (newRenamerRight, newProcess, rDefsLeft ++ rDefsRight)
       where
       newProcess   = ComponentSum newLeft newRight
       (newRenamerLeft,
        newLeft,
        rDefsLeft)  = liftProcess renamer left
       (newRenamerRight,
        newRight,
        rDefsRight) = liftProcess newRenamerLeft right
   -- Parallel processes should not really require this as they
   -- shouldn't contain any transitions, but we may later change
   -- our minds about that, and this doesn't hurt.
   liftProcess renamer (Cooperation left actions right)                      =
       (newRenamerRight, newProcess, rDefsLeft ++ rDefsRight)
       where
       newProcess   = Cooperation newLeft actions newRight
       (newRenamerLeft,
        newLeft,
        rDefsLeft)  = liftProcess renamer left
       (newRenamerRight,
        newRight,
        rDefsRight) = liftProcess newRenamerLeft right
   liftProcess renamer proc@(ProcessArray _ _ _)                             =
       (renamer, proc, [])
   liftProcess renamer (Hiding proc coopActs)                                =
       (newRenamer, newProc, newRDefs)
       where
       newProc      = Hiding newNextP coopActs
       (newRenamer,
        newNextP,
        newRDefs)   = liftProcess renamer proc



{- 
   Remove parallel definitions
   So this is a bit of a misnomer, it doesn't remove the parallel
   definitions, but it does expand them in the main system component.
   Perhaps it should then march down the process definitions and
   remove them from the list.
-}
removeParallelDefs :: ParsedModel -> ParsedModel
removeParallelDefs (ParsedModel rDefs pDefs system) =
   ParsedModel rDefs pDefs newSystem
   where
   newSystem = expandParallelAliases system

   -- Again here if the process is inherently a sequential process
   -- we do not bother recursing to the sub-process, for example
   -- prefix components we do not recurse to the next process
   -- since this should not contain any parallel components 
   -- (idents or otherwise)
   expandParallelAliases :: ParsedComponent -> ParsedComponent
   expandParallelAliases (StopProcess)     =
      StopProcess
   expandParallelAliases (IdProcess ident) =
       case expandIdent [] ident of
           Nothing -> IdProcess ident
           Just p  -> p
   expandParallelAliases proc@(PrefixComponent _ _)    = proc
   expandParallelAliases proc@(ComponentSum _ _)       = proc
   expandParallelAliases proc@(CondBehaviour _ _)      = proc
   expandParallelAliases (Cooperation left acts right) =
       Cooperation (expandParallelAliases left) acts
                   (expandParallelAliases right)
   expandParallelAliases proc@(ProcessArray _ _ _)     = proc
   expandParallelAliases (Hiding proc acts)            =
       Hiding (expandParallelAliases proc) acts

   expandIdent :: [ ParsedComponentId ] -> ParsedComponentId
               -> Maybe ParsedComponent
   expandIdent seen ident
       -- Not our job to deal with alias loops.
       | elem ident seen = Nothing
       | otherwise       =
           case lookup ident pDefs of
               -- Shouldn't really happen, undefined process
               Nothing                       -> Nothing
               Just (IdProcess ident')       -> 
                   expandIdent (ident : seen) ident'
               p@(Just (Cooperation _ _ _))  -> p
               p@(Just (ProcessArray _ _ _)) -> p
               p@(Just (Hiding _ _))         -> p
               -- finally if it is a sequential component leave it as is
               Just _                        -> Nothing


{-| 
   Remove needless aliases.
   Sometimes it is convenient to write :
   @ P = S ; @
   This tranformation replaces every occurrence of @P@ in the model
   with a @S@.
   We must be careful to avoid cyclic replacements ie if we have
   @ P = S ;
     S = P
   @
   then we must at least do the right thing here. The static-analysis
   phase should catch this case, but the user may have suppressed static
   analysis.

   NOTE: Care required; this does not remove the alias definitions
   only the need for them. They can be removed with
   'removeUnusedDefinitions'
-}
removeAliases :: ParsedModel -> ParsedModel
removeAliases (ParsedModel rDefs pDefs model) =
   ParsedModel rDefs newPdefs $ expandProcess model
   where
   newPdefs = map (mapSnd expandProcess) pDefs
   -- First build up a list mapping aliases to their unaliased
   -- definition.
   aliasMap :: [ (ParsedComponentId, ParsedComponentId) ]
   aliasMap = mapMaybe (expandAliasPair . fst) pDefs

   expandAliasPair :: ParsedComponentId
                   -> Maybe (ParsedComponentId, ParsedComponentId)
   expandAliasPair ident =
       do ident' <- expandAlias [] ident
          return $ (ident, ident')

   expandAlias :: [ ParsedComponentId ] -> ParsedComponentId
               -> Maybe ParsedComponentId
   expandAlias seen ident
       -- It is not our job to find alias loops
       | elem ident seen = Nothing
       | otherwise       =
           -- So basically we want to look up the identifier
           -- in the process definitions, if it returns another
           -- alias then we look that up, if not then we return
           -- the given identifier since it's the end of an alias
           -- chain.
           case lookup ident pDefs of
               -- It's also not our job to find undefined processes
               Nothing                 -> Just ident
               Just (IdProcess ident') -> expandAlias (ident : seen) ident'
               Just _                  -> Just ident

   expandProcess :: ParsedComponent -> ParsedComponent
   expandProcess (StopProcess)                    =
      StopProcess
   expandProcess (IdProcess ident)                =
       case lookup ident aliasMap of
           Nothing -> IdProcess ident
           Just i  -> IdProcess i
   expandProcess (PrefixComponent t nextP)        =
       PrefixComponent t $ expandProcess nextP
   expandProcess (ComponentSum left right)        =
       ComponentSum (expandProcess left) (expandProcess right)
   expandProcess (CondBehaviour cond nextP)       =
      CondBehaviour cond $ expandProcess nextP
   expandProcess (Cooperation left actions right) =
       Cooperation (expandProcess left) actions (expandProcess right)
   expandProcess proc@(ProcessArray _ _ _)        =
       proc
   expandProcess (Hiding proc actions)            =
       Hiding (expandProcess proc) actions



{-
   Remove process arrays.
   This tranformation removes process arrays, one of two things can be done.
   Either a process such as
   @ P[3] @
   can be turned into
   @ P <> P <> P @
   or we leave it as simply
   @ P @
   but return a concentration database.
-}
expandProcessArrays :: ParsedModel -> ParsedModel
expandProcessArrays (ParsedModel rDefs pDefs mainCoop) =
   ParsedModel rDefs newPdefs (expandProcess mainCoop)
   where
   newPdefs = map (mapSnd expandProcess) pDefs
   -- Notice for example that we do not recursively expand
   -- prefix components, since they should not contain
   -- parallel components. The same for component sums.
   expandProcess :: ParsedComponent -> ParsedComponent
   expandProcess proc@(IdProcess _)               = proc
   expandProcess proc@(StopProcess)               = proc
   {- Okay these three all have sub-processes which could
      be processed recursively but since they are all sequential
      processes they should not have any process arrays within
      them.
   -}
   expandProcess proc@(PrefixComponent _ _)       = proc
   expandProcess proc@(ComponentSum _ _)          = proc
   expandProcess proc@(CondBehaviour _ _)         = proc
   expandProcess (Cooperation left actions right) =
       Cooperation (expandProcess left) actions (expandProcess right)
   expandProcess (ProcessArray ident i mActions)
       | i < 1     = error "Process array with less than 1"
       | otherwise = expandArray (fromInteger i) ident $ maybe [] id mActions
   expandProcess (Hiding proc actions)              =
       Hiding (expandProcess proc) actions

   expandArray :: Int -> ParsedComponentId -> [ ParsedAction ]
               -> ParsedComponent
   expandArray x ident actions
       | x < 1     = IdProcess ident
       | otherwise = 
           Cooperation (IdProcess ident) actions 
               (expandArray (x - 1) ident actions)

{-|
   Remove the conditional behaviour based upon the values of rates
   within the model. This should clearly be applied after any rates
   have been overridden or selected from a set definition.

   TODO: we should reduce the rate expressions such that they are
   all simple rate numbers, but for now I will leave that for later.
-}
removeConditionalBehaviour :: ParsedModel -> ParsedModel
removeConditionalBehaviour (ParsedModel rateDefs processDefs mainCoop) =
   ParsedModel rateDefs newProcessDefs $ removeFromProcess mainCoop
   where
   newProcessDefs = map (mapSnd removeFromProcess) processDefs

   removeFromProcess :: ParsedComponent -> ParsedComponent
   removeFromProcess (CondBehaviour cond next)
      | condIsTrue cond = removeFromProcess next
      | otherwise       = StopProcess
   removeFromProcess (ComponentSum left right)     =
      case (newLeft, newRight) of
         (StopProcess, _) -> newRight
         (_, StopProcess) -> newLeft
         _                -> ComponentSum newLeft newRight
      where
      newLeft  = removeFromProcess left
      newRight = removeFromProcess right
   removeFromProcess p@(IdProcess _)               = p
   removeFromProcess p@(StopProcess)               = p
   removeFromProcess (PrefixComponent trans next)  =
      PrefixComponent trans $ removeFromProcess next
   removeFromProcess (Cooperation left acts right) =
      Cooperation (removeFromProcess left)
                  acts
                  (removeFromProcess right)
   removeFromProcess p@(ProcessArray _ _ _)        = p
   removeFromProcess (Hiding comp actions)         =
      Hiding (removeFromProcess comp) actions 

   -- Obviously this is not quite sophisticated enough
   condIsTrue :: BeCondition -> Bool
   condIsTrue (BeCondition rateId operator rateExp) =
      case (lookup rateId rateDefs, rateExp) of
         (Just (RateNum v1), RateNum v2) -> interpretOperator operator v1 v2
         (Nothing, _)                    -> 
            error "undefined rate identifier in conditional behaviour"
         (_, _)       ->
            error "unimplemented rate expressions in a conditional bevahiour"

   interpretOperator :: String -> ParsedValue -> ParsedValue -> Bool
   interpretOperator "<" = (<)
   interpretOperator ">" = (>)
   interpretOperator _   = 
      error "Unimplemented conditional behaviour operator"


{-|
   This function performs a similar function to the previous
   'removeConditionalBehaviour' but it may be used ahead of the
   instantiation of rate identifiers. It moves the conditional
   behaviour inside the /ParsedTrans/ by making it a functional
   rate which is either the normal value or zero depending on
   the condition.
   Note that this function only works if the conditional
   behaviour is a prefix operator.

   In the other behaviour we could conceivably create a whole
   new process but for now we simply reject any other attempt.
-}
relegateConditionalBehaviour :: ParsedModel -> ParsedModel
relegateConditionalBehaviour (ParsedModel rateDefs processDefs mainCoop) =
   ParsedModel rateDefs newProcessDefs $ removeFromProcess mainCoop
   where
   newProcessDefs = map (mapSnd removeFromProcess) processDefs

   removeFromProcess :: ParsedComponent -> ParsedComponent
   removeFromProcess (CondBehaviour (BeCondition rateId operator rExp)
                                    (PrefixComponent trans next) ) =
      case pepaTransRate trans of
         RateTimed rateExp ->
            PrefixComponent newTrans $ removeFromProcess next
            where
            newTrans = trans { pepaTransRate = RateTimed newRate }
            newRate  = RateCond cond rateExp $ RateNum 0.0
            cond     = PCondRate rateId operator rExp
         _                 ->
            error "relegateConditionalBehaviour, unimplemented rate"
   removeFromProcess (CondBehaviour _ _)           =
      error "relegateConditionalBehaviour, unimplemented condition"
   removeFromProcess (ComponentSum left right)     =
      case (newLeft, newRight) of
         (StopProcess, _) -> newRight
         (_, StopProcess) -> newLeft
         _                -> ComponentSum newLeft newRight
      where
      newLeft  = removeFromProcess left
      newRight = removeFromProcess right
   removeFromProcess p@(IdProcess _)               = p
   removeFromProcess p@(StopProcess)               = p
   removeFromProcess (PrefixComponent trans next)  =
      PrefixComponent trans $ removeFromProcess next
   removeFromProcess (Cooperation left acts right) =
      Cooperation (removeFromProcess left)
                  acts
                  (removeFromProcess right)
   removeFromProcess p@(ProcessArray _ _ _)        = p
   removeFromProcess (Hiding comp actions)         =
      Hiding (removeFromProcess comp) actions 

{-|
   Remove unused definitions. This should obviously occur after we have checked
   for unused definitions otherwise the type checker \/ analyser will not pick
   this up. However for example the srmc compiler must select process
   definitions and therefore we do not wish to compile out unnecessary definitions.
   In particular as well rate arrays should not be ranged over if the rate
   is never used.

   Rather than calling 'Language.Pepa.Analysis.Analysis.analyseModel' ourselves
   this function takes in an analysis report. This is in general the correct
   thing to do because there may be multiple functions which use an analysis
   report and we do not wish to compute it several times. Additionally here
   we wish to make sure that analysis has already occurred and accepting the
   report as an argument ensure that at least the analysis has occurred if
   maybe the results have been ignored.

   ARGH: unfortunately the current analysis only tells whether a rate is used
   based on whether or not it appears in the model and *not* whether or not
   that process is used. Hence unfortunately we will have to re-analyse
   the model after removing the unused processes, sucks but then, what
   can you do?
-}
removeUnusedDefinitions :: AnalysisReport -> ParsedModel -> ParsedModel
removeUnusedDefinitions report (ParsedModel rDefs pDefs mainComp) =
   ParsedModel newRateDefs newProcessDefs mainComp
   where
   -- The interModel is the original model with the unused process
   -- definitions removed. We must re-analyse this to find out which
   -- rates are not used.
   interModel = ParsedModel rDefs newProcessDefs mainComp
   pNames     = usedProcessNames report

   reReport   = analyseModel interModel
   rNames     = usedRateNames reReport

   newProcessDefs = filter ((flip elem pNames) . fst) pDefs
   newRateDefs    = filter ((flip elem rNames) . fst) rDefs


{- The type used to record which process definitions we have "seen" -}
type Seen = [ ParsedComponentId ]

{-|
   The idea of qualifying a model is to take a model and make unique
   every occurrence of a process identifier in the main composition.
   So for example something like, @ P <> P @, becomes
   @ P_1 <> P_2 @.
   This has consequences for later looking it up in the process
   definition list of course.
   
   Note: once we re-introduce parallel definitions, we will have to
   work out how to deal with those, perhaps we just compile them
   out before hand?
   
   Also I think it would be desirable to leave as unqualified the first
   occurrence so that in the case that there is only one occcurrence of
   any process we do not do any qualification.

   It is important to note that there are one or two functional
   dependencies on this model. First of all there are no aliases
   or parallel definitions. Therefore any used process name we
   can assume is a sequential definition.
-}
qualifyModel :: ParsedModel -> ParsedModel
qualifyModel (ParsedModel rateDefs processDefs pComp) =
   ParsedModel rateDefs qualifiedDefs newComp
   where
   (_id, newComp, qualifiedDefs) = qualifyPar 0 pComp

   qualifyPar :: Int -> ParsedComponent
	       -> (Int, ParsedComponent, [ ProcessDef ])
   qualifyPar idNum (ProcessArray ident size mActions) =
      (idNum + 1, newPar, newDefs)
      where
      (_seen, newDefs) = qualifyProcess idNum [ ident ] ident
      newIdent         = qualifyQName ident idNum
      newPar           = ProcessArray newIdent size mActions 
   qualifyPar idNum (IdProcess ident)                  =
      (idNum + 1, newPar, newDefs)
      where
      (_seen, newDefs) = qualifyProcess idNum [ ident ] ident
      newIdent         = qualifyQName ident idNum
      newPar           = IdProcess newIdent
   qualifyPar idNum (Cooperation left actions right)   =
      (idNum2, newCoop, leftDefs ++ rightDefs)
      where
      newCoop  = Cooperation newLeft actions newRight
      (idNum1, newLeft, leftDefs)   = qualifyPar idNum left
      (idNum2, newRight, rightDefs) = qualifyPar idNum1 right
   qualifyPar idNum (Hiding hComp actions)              =
      (idNum1, Hiding newHComp actions, newDefs)
      where
      (idNum1,  newHComp, newDefs) = qualifyPar idNum hComp
   qualifyPar _idNum StopProcess                        =
      error "Sequential component found in parallel composition"
   qualifyPar _idNum (PrefixComponent _ _)              =
      error "Sequential component found in parallel composition"
   qualifyPar _idNum (ComponentSum _ _)                 =
      error "Sequential component found in parallel composition"
   qualifyPar _idNum (CondBehaviour _ _)                =
      error "Sequential component found in parallel composition"   



    {-
      Qualify process must search in the *original* definition list
      for the name we are qualifying and return the list of
      qualified definitions. It doesn't need to return the list
      of seen process names (*I think*) because because if there
      is a recursive definition the name is added before we do
      the rest of the definition.

      Also note that it doesn't *generate* names, this is because
      all we are doing is applying the current renaming to the
      processes. However once parallel definitions are allowed this
      would change, since in the case of a parallel definition the
      process identifiers contained therein would have to be given
      new names (as they are essentially just new occurences within
      the main composition).
    -}
   qualifyProcess :: Int -> Seen -> ParsedComponentId 
                 -> (Seen, [ ProcessDef ])
   qualifyProcess idNum seen ident =
      (newSeen, newProcessDef : newDefs)
      where 
      newProcessDef               = (newIdent, newSequ)
      newIdent                    = qualifyQName ident idNum
      (newSeen, newSequ, newDefs) = qualifySeqWith idNum seen process
      process                     = fromMaybe err mProcess
      mProcess                    = lookup ident processDefs
      err  = error "There should be no undefined processes"



    {-
      Qualifying a definition does not return an integer, because
      we want to qualify all the names within the definition with
      the same integer. We are not at this point introducing names.

      Note: this would not be the case for parallel definitions.
      
      We require the integer with which to qualify all of the names
      and also the names of the definitions we have already seen
      in order that we do not go into a loop.

      Note also that we are currently not doing anything with rate
      definitions, but since these should ultimately contain
      process names within functional rates we definitely should.
      *What* we should do is not entirely obvious.
    -}
 
   qualifySeqWith :: Int -> [ParsedComponentId] -> ParsedComponent
                  -> (Seen, ParsedComponent, [ ProcessDef ])
   qualifySeqWith idNum seen (IdProcess ident)
      | elem ident seen = ( seen, sequ, [])
      | otherwise       = ( newSeen, sequ, newDefs )
      where
      sequ               = IdProcess $ qualifyQName ident idNum
      tmpSeen            = ident : seen
      (newSeen, newDefs) = qualifyProcess idNum tmpSeen ident
   qualifySeqWith idNum seen (PrefixComponent trans sequ) =
      (newSeen, newPrefix, newDefs)
      where
      {- Note here we do nothing with the trans, might have to
      when we introduce functional rates -}
      (newSeen, newSequ, newDefs) = qualifySeqWith idNum seen sequ
      newPrefix = PrefixComponent trans newSequ
   qualifySeqWith idNum seen (ComponentSum left right) =
      (seenRight, newSum, leftDefs ++ rightDefs)
      where 
      newSum = ComponentSum newLeft newRight
      (seenLeft,  newLeft,  leftDefs)  = qualifySeqWith idNum seen left
      (seenRight, newRight, rightDefs) = qualifySeqWith idNum seenLeft right

   qualifySeqWith _idNum seen proc@StopProcess             =
      (seen, proc, [])
   qualifySeqWith idNum seen (CondBehaviour cond proc)     =
      (newSeen, CondBehaviour cond newP, newDefs)
      where
      (newSeen, newP, newDefs) = qualifySeqWith idNum seen proc
   qualifySeqWith _idNum _seen (Cooperation _ _ _)         =
      error "parallel definition found while qualifying the model"
   qualifySeqWith _idNum _seen (Hiding _ _ )               =
      error "parallel definition found while qualifying the model"
   qualifySeqWith _idNum _seen (ProcessArray _ _ _)        =
      error "parallel definition found while qualifying the model"

{- The type of the action map, mapping action identifiers to new identifiers -}
type Mapping a b = [ (a, b) ]
type ActionMap = Mapping ActionIdentifier ActionIdentifier

{-|
   The next transformation removes hiding from the model by renaming
   the actions. Note that this must be performed on a model which
   has already been qualified.

   For this reason we provide a function to remove hiding from an unqualifed
   model. Note that this simplfy calls 'qualifyModel' first.
   Perhaps later if we wish to do pepa->pepa translations which remove
   hiding then we can worry about doing this while only qualifying what
   is necessary.
-}
removeHidingUnqualifiedModel :: ParsedModel -> ParsedModel
removeHidingUnqualifiedModel = removeHiding . qualifyModel


{-|
   Removing hiding from a qualified model. This if fairly simple
   we need only rename those actions which are hidden we do not
   need to create new definitions only modify current ones.
   Since each definition is only used by one process so the actions
   which may be performed by each defined process are either hidden
   or not. (Not for example hidden in one use and not in another.)
-}
removeHiding :: ParsedModel -> ParsedModel
removeHiding (ParsedModel rateDefs processDefs mainComp) =
   (ParsedModel rateDefs hiddenDefs newMainComp)
   where
   (_ns, hiddenDefs, newMainComp) = hideComp initialNameSupply [] [] mainComp

   hideComp :: NameSupply -> ActionMap -> Seen -> ParsedComponent
            -> ( NameSupply, [ ProcessDef ], ParsedComponent )
   hideComp names _aMap _seen proc@StopProcess                 =
      (names, [], proc)
   hideComp names aMap seen proc@(IdProcess ident)
      | elem ident seen = (names, [], proc)
      | otherwise       = 
         case lookup ident processDefs of
            Nothing -> error "Undefined process detected while removing hiding"
            Just p  -> 
               ( newNs, newDef : newDefs, proc)
               where
               newDef                  = (ident, newP)
               (newNs, newDefs, newP) = hideComp names aMap (ident : seen) p

   hideComp names aMap seen (PrefixComponent trans next)       =
      (newNs, newDefs, newPrefix)
      where
      (newNs, newDefs, newNext) = hideComp names aMap seen next
      newPrefix                  = PrefixComponent newTrans newNext
      newTrans                   = trans { pepaTransAction = newAction }
      newAction                  = hideAction aMap $ pepaTransAction trans
   hideComp names aMap seen (ComponentSum left right)          =
      ( newNsRight
      , unionBy equalDefs leftDefs rightDefs
      , ComponentSum newLeft newRight
      )
      where
      (newNsLeft, leftDefs, newLeft)    = hideComp names aMap seen left
      (newNsRight, rightDefs, newRight) = hideComp newNsLeft aMap seen right

      -- Although the model has been qualified a component sum may
      -- still have the same component in either half, consider
      -- @ P = (a, r).Q + (b, r) . Q @
      -- if we did the simple thing of hiding the left and right of the
      -- component sum we would duplicate the definitions for Q.
      -- Hence the final result above uses 'unionBy' and here we just
      -- define to definitions as equal if they defined the same process.
      equalDefs :: ProcessDef -> ProcessDef -> Bool
      equalDefs def1 def2 = (fst def1) == (fst def2)

   hideComp names aMap seen (Cooperation left actions right)   =
      ( newNsRight
      , leftDefs ++ rightDefs
      , Cooperation newLeft newActions newRight
      )
      where
      newActions = map (lookupParsedAction aMap) actions
      (newNsLeft, leftDefs, newLeft)    = hideComp names aMap seen left
      (newNsRight, rightDefs, newRight) = hideComp newNsLeft aMap seen right
   hideComp names aMap seen (ProcessArray ident size actions)  =
      case proc of
         IdProcess ident2 -> if ident == ident2
                             then (newNS, newDefs, newP)
                             else errMsg
         _                -> errMsg
      where
      errMsg = error "internal error recursive call of hideComp"
      newP                   = ProcessArray ident size newActions
      (newNS, newDefs, proc) = hideComp names aMap seen (IdProcess ident)
      newActions             = 
         do acts <- actions
            return $ map (lookupParsedAction aMap) acts
   hideComp names aMap seen (CondBehaviour cond proc)          =
      (newNs, newDefs, CondBehaviour cond newP)
      where
      (newNs, newDefs, newP) = hideComp names aMap seen proc
   {- The case for actual hiding -}
   hideComp names aMap seen (Hiding comp actions)              =
      hideComp newNames newActionMap seen comp
      where
      newActionMap    = snd newNamesAndAMap
      newNames        = fst newNamesAndAMap
      newNamesAndAMap =foldl addParsedAction (names, aMap) actions
      addParsedAction :: (NameSupply, ActionMap) -> ParsedAction
                      -> (NameSupply, ActionMap)
      addParsedAction (nSupply, m) (Action ident)    = 
         (newSupply, newM)
         where
         newM                 = (ident, newIdent) : m
         newIdent             = Unqualified newName
         (newSupply, newName) = getName nSupply $ "tau__" ++ (showOrig ident)
      -- Currently com actions are treated the same, but we may wish to
      -- say that you shouldn't be hiding a communication action??
      addParsedAction (nSupply, m) (ComAction ident) = 
         (newSupply, newM)
         where
         newM                 = (ident, newIdent) : m
         newIdent             = Unqualified newName
         (newSupply, newName) = getName nSupply $ "tauCA__" ++ (showOrig ident)
      addParsedAction _m (Tau _ )                    =
         error "A tau action in a hide set? shouldn't happen"

   {- These functions are merely for looking up the actions in an
      action map and are essentially just wrappers around the action
      data types (and the maybe).
   -}
   hideAction :: ActionMap -> ParsedAction -> ParsedAction
   hideAction m (Action    ident) = Action   $ lookupAction ident m
   hideAction m (Tau       ident) = Tau      $ lookupAction ident m
   hideAction m (ComAction ident) = ComAction $ lookupAction ident m

   lookupParsedAction :: ActionMap -> ParsedAction -> ParsedAction
   lookupParsedAction actionMap (Action    ident) = 
      Action    $ lookupAction ident actionMap
   lookupParsedAction actionMap (Tau       ident) =
      Tau       $ lookupAction ident actionMap
   lookupParsedAction actionMap (ComAction ident) =
      ComAction $ lookupAction ident actionMap

   lookupAction :: ActionIdentifier -> ActionMap -> ActionIdentifier
   lookupAction ident mapping = 
    fromMaybe ident $ lookup ident mapping
