{-|
  This is a utilities module containing some functions that are used
  throughout ipclib.
 -}
module Language.Pepa.Utils 
   ( fstTriple
   , sndTriple
   , trdTriple
   , selectTerm
   , replaceTerm
   , nSpace
   , (?), (==>)
   , findIntersectionBy
   , showCol
   , showTable
   , showTableVarCol
   , ColWidth
   , ColJustify(..)
   , choose
   , factorial
   , divFactDouble
   , fromLeft
   , fromRight
   , isLeft
   , isRight
   , fromMaybeF
   , throughMaybe
   , findInPair
   , findInPairErr
   , getPartitions
   , partitionTwo
   , powerSet
   , indent
   , mapFst
   , mapSnd
   , numberLines
   , fromInt
   , isListUniform
   , permuteLists
   , permuteDefinitionArrays
   , splitFileExt
   , removeSuffix
   , switchSuffix
   , splitFilePath
   , dirName
   , switchDirectory
   , switchBaseName
   , addLeadingDotSlash
   , mkCSlist
   , mkSepList
   , mkGenCSlist
   , mkGenSeplist
   , fullGroupBy
   , dropFromEnd
   )
where


{- External Library Imports -}
{- Standard Library Imports -}
import Data.Maybe
   ( isNothing
   , fromJust 
   )
import Data.List
   ( find
   , transpose
   , (\\)
   , partition
   )
{- Local module imports -}
{- End of Imports -}


-- <GHCi compatability mode>
fromInt :: (Num b, Integral a) => a -> b
fromInt = fromIntegral
-- </GHCi compatability mode>


{-|
  These 'mapFst' and 'mapSnd' are actually defined in
  @Data.Graph.Inductive.Query.Monad@, no idea why they are defined
  in such an obscure path, anyway to avoid importing a whole package
  (@fgl@) just for these to functions I just redefine them here.
-}
mapFst :: (a -> b) -> (a, c) -> (b, c)
mapFst f (a, c) = (f a, c)

{-| Apply a function to the second component of a pair. -}
mapSnd :: (a -> b) -> (c, a) -> (c, b)
mapSnd f (c, a) = (c, f a)


{-| Add numbers to a list of strings -}
numberLines :: [String] -> String
numberLines ss = 
    {-
      @spaces@ is the amount of white space required to add to the
      end of the number so that all of the lines start at the same
      indentation. So if there are @100@ lines then we need two
      spaces for the single figure digits and one for the double
      and none for line @100@.
    -}
    unlines [ let spaces = take amount  $ repeat ' '
	          amount = (length $ show $ length ss) - (length $ show i)
	      in concat [spaces, show i," : ", s] | (i, s) <- zip ints ss ]
    where ints   :: [ Integer ]
	  ints   = [1..]


fstTriple :: (a, b, c) -> a
fstTriple (a, _, _ ) = a

sndTriple :: (a, b, c) -> b
sndTriple (_, b, _) = b

trdTriple :: (a, b, c) -> c
trdTriple (_, _, c) = c


-- return correct intersection between two lists (ie multisets)
findIntersectionBy :: (a -> a -> Bool) -> [a] -> [a] -> [a]
findIntersectionBy _ [] _  = []
findIntersectionBy f (r:rs) ss 
    | isNothing mf = findIntersectionBy f rs ss
    | otherwise    = (r : findIntersectionBy f rs ss') where
		  mf = selectTerm (f r) ss
		  (_, ss') = fromJust mf

-- selects the first item from a list (if it exists) satisfying f and presents the item and the list\item
selectTerm :: (a -> Bool) -> [a] -> Maybe (a, [a])
selectTerm f ys
	| null ys2	= Nothing   
	| otherwise	= Just (head ys2, ys1 ++ tail ys2) where
                                (ys1, ys2) = span (not.f) ys

-- replaces elems in list xs that pass f with y
replaceTerm :: (a -> Bool) -> a -> [a] -> [a]
replaceTerm f y xs = map (\x -> f x ? (y,x)) xs 

-- finds b in [(a,b)] by using index element a
findInPair :: Eq a => a -> [(a,b)] -> Maybe b
findInPair z ys = isNothing mfd ? (Nothing, Just $ snd $ fromJust mfd) where
		  mfd = find (\x -> fst x == z) ys
-- as above but produces an error if not found
findInPairErr :: Eq a => (String,String) -> a -> [(a,b)] -> b
findInPairErr (fn,mess) x ys
    | isNothing mfd = error (fn ++ ": " ++ mess)
    | otherwise = fromJust mfd where
		  mfd = findInPair x ys


isLeft :: Either a b -> Bool
isLeft (Left _) = True
isLeft _ = False

isRight :: Either a b -> Bool
isRight = not.isLeft

fromLeft :: Either a b -> a
fromLeft (Left x) = x
fromLeft _        = error "attempt to take a Left from a Right"

fromRight :: Either a b -> b
fromRight (Right y) = y
fromRight _         = error "attempt to take a Right from a Left"

(?) :: Bool -> (a, a) -> a
(?) bl e = if bl then (fst e) else (snd e)

{-
  [-fromMaybeF-] if there is a value there the argument function is applied
  to it, if not then the default value is returned.
-}
fromMaybeF :: (a -> b) -> b -> Maybe a -> b
fromMaybeF f _ (Just a) = f a
fromMaybeF _ b Nothing  = b

{-|
   This function performs a given transformation inside of the 'Maybe' data type.
   This could actually be defined using the monad operators.
   For those unfamiliar with monad operators this just returns @Nothing@ in the
   case that is is given @Nothing@, and @Just (f a)@ in the case that it is given
   @Just a@.
-}
throughMaybe :: (a -> b) -> Maybe a -> Maybe b
throughMaybe f m = m >>= (return . f)

(==>) :: Bool -> Bool -> Bool
(==>) a b = (not a) || b

-- String Related

nSpace :: Int -> String
nSpace n
	| n < 0		= "" -- error "Utils.nSpace: negative argument"
	| otherwise	= concat $ take n $ repeat " "


-- Maths funclets
{-
permutations :: [a] -> [[a]]
permutations xs = map (\(y,ys) -> y:ys) $ getPartitions xs
-}
getPartitions :: [a] -> [(a, [a])]
getPartitions ls = partitions ls [] where
	partitions :: [a] -> [a] -> [(a, [a])]
	partitions [] _ = []
	partitions (x:xs) ys = (x, ys ++ xs) : partitions xs (x:ys)

factorial :: Int -> Int
factorial  0 = 1
factorial  1 = 1
factorial  2 = 2
factorial  3 = 6
factorial  4 = 24
factorial  5 = 120
factorial  6 = 720
factorial  7 = 5040
factorial  8 = 40320
factorial  9 = 362880
factorial 10 = 3628800
factorial 11 = 39916800
factorial 12 = 479001600
factorial n  = error ("Utils.factorial: Int overflow for " ++ (show n) ++ "!")

divFactorial :: Integer -> Int -> Integer
divFactorial p q 
	| q < 13	= div p (fromInt $ factorial q)
	| otherwise	= foldl div p $ map fromInt [1..q]

divFactDouble :: Double -> Int -> Double
divFactDouble p q
	| q < 13	= p / (fromIntegral $ factorial q)
	| otherwise	= foldl (/) p $ map fromInt [1..q]

choose  :: Int -> Int -> Integer
choose n r
	| r > n		= error "Utils.choose: r > n in nCr"
	| n == r	= 1
	| n < 13	= fromInt $ div (div (factorial n) (factorial $ n - r)) (factorial r)
	| otherwise	= (2*r < n) ? (f n r, f n (n-r)) where
		f = \x -> \y -> divFactorial (product $ g [x - y + 1..x]) y
		g = \xs -> map fromInt xs

indent :: String -> String
indent inset = unlines $ map ('\t':) (lines inset)

-- Column display routines

data ColJustify = L | C | R
type ColWidth = Int

-- If string s exceeds column width then it is allowed to overflow
showCol :: ColWidth -> ColJustify -> String -> String
showCol w j s
	| length s > w		= case j of
		R	-> (w > 0) ? ("$" ++ (drop (1 + (length s) - w) s), "")
		_	-> (w > 0) ? ((take (w - 1) s) ++ "$", "")
	| otherwise		= case j of
		L	-> s ++ (nSpace $ w - length s)
		C	-> sp1 ++ s ++ sp2 where
				sp = w - length s
				sp1 = nSpace (div sp 2)
				sp2 = (mod sp 2 == 1) ? (sp1 ++ " ", sp1)  
		R	-> (nSpace $ w - length s) ++ s


showTableVarCol :: [[String]] -> String
showTableVarCol xss = showTable cs xss where
	ns = map maximum $ map (map length) (transpose xss)
	cs = [((1,""),n,L) | n <- ns]

-- 1st cw param indicates separation between cols, 2nd is width of column itself
showTable :: [((ColWidth, String), ColWidth, ColJustify)] -> [[String]] -> String
showTable _ [] = ""
showTable forms (xs:xss) = (showRow $ zip forms xs) ++ (showTable forms xss) where
	showRow :: [(((ColWidth, String), ColWidth, ColJustify), String)] -> String
	showRow [] = "\n"
	showRow ((((ws, cdv), wc, just), str):cs) = (showCol ws C cdv) ++ (showCol wc just str) ++ (showRow cs)


-- Create comma separated list of a list of generic showable type
-- You might wish to change his to have  space between each item if you
-- you do, remember that the condor arguments outputs a rate array as
-- a comma separated list, and it *cannot* have spaces in it.
mkGenCSlist :: Show a => [a] -> String
mkGenCSlist [] = ""
mkGenCSlist xs = foldr1 (\x -> \y -> x ++ "," ++ y) (map show xs)

-- Create a separated list of showable items with a custom separator
mkGenSeplist :: Show a => String -> [a] -> String
mkGenSeplist _ []   = ""
mkGenSeplist sep xs = foldr1 (\x -> \y -> x ++ sep ++ y) (map show xs)


-- Create comma separated list of a string list
-- see above comment for mkGenCSlist
mkCSlist :: [ String ] -> String
mkCSlist = mkSepList ","

{-| Create a list separated by the given separator -}
mkSepList :: String -> [ String ] -> String
mkSepList _sep [] = ""
mkSepList sep  l  =  foldr1 (\x -> \y -> x ++ sep ++ y) l

-- Generates power set of a list of elements
powerSet :: [a] -> [[a]]
powerSet []     = [[]]
powerSet (x:xs) = [ x : ys | ys <- powerSet xs ] ++ (powerSet xs)

-- Creates every possible two-partition of a list of elements
partitionTwo :: Eq a => [a] -> [([a], [a])]
partitionTwo xs = [ (ys, xs \\ ys) | ys <- powerSet xs ]

{-
  Checks if all the elements of a list share a common property.
  For example if it is a list of lists then we can check if they all have
  the same length.
-}
isListUniform :: Eq b => (a -> b) -> [a] -> Bool
isListUniform _ []    = True -- Always true for the empty list
isListUniform _ [ _ ] = True -- Always true for the singleton list
isListUniform f (first : second : rest)
    | f first == f second = isListUniform f (second : rest)
    | otherwise           = False

{-
  Permutates a list of lists. Essentially this multiplies a list of lists
  together. We return a list of lists, all of the lists returned will be
  the same length. Each index into all of the lists will be one permutation
  of the values in the original lists. So the first element of the first
  list input list will, in the returned list of lists, share a row with all
  of the elements in each of the other lists.
  Eg, suppose we are given the lists 
  @
  [1, 2, 3]
  [4, 5 ]
  [7]
  @
  Then the returned output is
  @
  [ 1, 1, 2, 2, 3, 3 ]
  [ 4, 5, 4, 5, 4, 5 ]
  [ 7, 7, 7, 7, 7, 7 ]
  @

  To do this, the first two cases are obvious, permuting the empty list of
  lists, or the singleton list of one list just returns the argument.
  Others, we must essentially multiply the first list with the permuted
  rest of the lists. The @permutedRest@ permutes the rest of the lists
  this should produce a list of lists all of the same size. We then repeat
  each of these the number of times equal to the length of the first list.
  The first list itself has each element repeated the number of times equal
  to the size of the original permuted lists for the rest. Essential then
  all lists should be equal to the length of the first list multiplied by
  the length of an individual list in the recursively permuted rest. Since
  all the lists in the recursively permuted rest are of equal length it
  doesn't matter which one, we take the first.
-}
permuteLists :: [[a]] -> [[a]]
permuteLists []             = []
permuteLists [ lastList ]   = [ lastList ]
permuteLists (first : rest) = 
   (newFirst : productPermed)
   where 
   permutedRest  = permuteLists rest
   productPermed = map (replicateList lengthFirst) permutedRest
   -- We can take the head of the first, since [-rest-] cannot
   -- be empty otherwise the first case would have applied.
   sizeRest      = length $ head permutedRest
   lengthFirst   = length first
   newFirst      = replicateEachElement sizeRest first


{-
   Permute set definitions, basically performs the above 'permuteLists'
   but rather than on a set of list, ie @[ [a] ]@ on a set of definition
   lists, ie a set of mappings from somethingt to lists, that is
   @ [ (a, [b]) ] @

   Note that this does rely on the above 'permuteLists' not
   changing the order of the lists.
-}
permuteDefinitionArrays :: [ (a, [b]) ] -> [ (a, [b]) ]
permuteDefinitionArrays arrays = 
   zip names newArrays
   where
   (names, shortArrays) = unzip arrays
   newArrays            = permuteLists shortArrays

{-
  Replicates a list a number of times, so for example 
  [- replicateList 3 [1,2,3] -] would be
  [- [1,2,3, 1,2,3, 1,2,3] -]. To do this we replicate the list
  [-n-] number of times to get a list of lists, and then concatenate them.
-}
replicateList :: Int -> [a] -> [a]
replicateList n l = concat $ replicate n l

{-
  Replicates each element in a list a number of times, so for example
  [- replicateEachElement 3 [1,2,3] -] returns
  [- [1,1,1, 2,2,2, 3,3,3] -].
-}
replicateEachElement :: Int -> [a] -> [a]
replicateEachElement _ []             = []
replicateEachElement n (first : rest) = 
    (replicate n first) ++ (replicateEachElement n rest)

{-
  \subsection{File Extensions}
  A couple of functions to operate over filename extensions, I sort of
  expected these to be in the standard libraries but I cannot find them
  if anyone knows them, please take these out and replace uses of them,
  but please leave a comment here stating where you found them.

  Update: There has been a discussion on the Haskell-Cafe mailing list
  and an announcement of a release of a library module
  [-System.FilePath-]. I think it is worth waiting until that is in
  a released version of ghc before switching over to it, but we
  then should do that.

  Alternatively just make these synonyms ie.
  [-removeSuffix = Some.Library.function-]
  I think it can actually be found in
  [-Distribution.Compat.FilePath-] but I don't seem to be allowed to
  import that module.

  The first of these [-splitSuffix-] returns the base name and the extension
  the dot is included as part of the extension.
-}
splitFileExt :: FilePath -> (String, String)
splitFileExt []              = ([], [])
splitFileExt ('.' : rest)
    | restSuf == [] = ([], '.' : restPre)
    | otherwise     = (('.' : restPre), restSuf)
    where (restPre, restSuf) = splitFileExt rest
splitFileExt (first : rest)  = 
    ((first : restPre) , restSuf)
    where (restPre, restSuf) = splitFileExt rest


{-
  Removes the given suffix if it is the suffix of the file path given,
  otherwise leaves the file path untouched.
-}
removeSuffix :: String -> FilePath -> FilePath
removeSuffix ext fileName
  | ext == fileExt = fileBase
  | otherwise      = fileName
  where (fileBase, fileExt) = splitFileExt fileName


{-
  Switches a suffix, this allows us to for example easily get the mod file
  from the pepa file, that is
  [-switchSuffix ".pepa" ".mod" "file.pepa" -]
  will evaluate to
  [-"file.mod"-]
-}
switchSuffix :: String -> String -> FilePath -> FilePath
switchSuffix currentSuffix desiredSuffix fileName =
    (removeSuffix currentSuffix fileName) ++ desiredSuffix


{-
  Splits the file path into directory name and base name of the file.
  Very similar to [-splitFileExt-] but that the trailing directory separator
  appended onto the end of the directory name rather than the base name.
-}
splitFilePath :: FilePath -> (FilePath, String)
splitFilePath fileName =
   {-
     The [-localSplit-] function is almost the right definition for
     this but it doesn't work in the case that there are no directory
     deliminators, that is if we a name such as [-"airbag.pepa"-].
     So we just check if that is the case and take appropriate action.
   -}
   case localSplit fileName of
   (s, []) -> ([], s)
   other   -> other
   where localSplit :: FilePath -> (FilePath, String)
	 localSplit []              = ([], [])
	 localSplit ('/' : rest)
	     | restSuf == [] = (['/'], restPre)
	     | otherwise     = (('/' : restPre), restSuf)
	     where (restPre, restSuf) = localSplit rest
	 localSplit (first : rest)  = 
	     ((first : restPre) , restSuf)
	     where (restPre, restSuf) = localSplit rest

{-|
  Returns the directory component of the given file path
-}
dirName :: FilePath -> FilePath
dirName = fst . splitFilePath


{-|
  Takes in a file and removes the directory path and appends the new directory
  path.
-}
switchDirectory :: FilePath -> FilePath -> FilePath
switchDirectory fileName newDir =
    newDir ++ "/" ++ baseName
    where (_, baseName) = splitFilePath fileName

{-|
   Takes in a filepath and a file name and replaces the basename part
   of the given filepath with the new filename
-}
switchBaseName :: FilePath -> FilePath -> FilePath
switchBaseName fileName newFile =
   (dirName fileName) ++ "/" ++ newFile

{-|
   Adds in a leading @.\/@ should the path not begin with a @\/@.
   This should really be from the @filepath@ library.
-}
addLeadingDotSlash :: FilePath -> FilePath
addLeadingDotSlash fn@('/' : _)       = fn
addLeadingDotSlash fn@('.' : '/' : _) = fn
addLeadingDotSlash fn                 = "./" ++ fn


{-
   fullGroupBy operates over the whole list at all stages, unlike
   'Data.List.groupBy'.
   Hence @groupBy (==) [1, 2, 2, 1] = [[1], [2, 2], [1]]@
   but   @fullGroupBy (==) [1, 2, 2, 1] = [[1, 1], [2, 2]]@
-}
fullGroupBy :: (a -> a -> Bool) -> [ a ] -> [ [ a ] ]
fullGroupBy _ []         = []
fullGroupBy f (h : rest) =
   (h : withH) : (fullGroupBy f notWithH)
   where
   (withH, notWithH) = partition (f h) rest

{-|
   Same as 'dropWhile' but drops from the end of the list.
-}   
dropFromEnd :: (a -> Bool) -> [a] -> [a]
dropFromEnd f =    reverse . (dropWhile f) . reverse